import datetime
import warnings

import numpy as np
from estimagic.differentiation.derivatives import first_derivative
from estimagic.exceptions import get_traceback
from estimagic.logging.database_utilities import append_row
from estimagic.optimization.process_results import switch_sign
from estimagic.utilities import hash_array


DERIVATIVE_ERROR_MESSAGE = (
    "Error during derivative evaluation at parameters at which the criterion has "
    "already been evaluated and used by the optimizer before. Thus it is not possible "
    "to simply replace the criterion function by a penalty function."
)


CRITERION_ERROR_MESSAGE = (
    "Error during criterion evaluation at parameters at which the derivative has "
    "already been evaluated and used by the optimizer before. Thus it is not possible "
    "to simply replace the criterion function by a penalty function."
)


NO_PRIMARY_MESSAGE = (
    "The primary criterion entry of the {} algorithm is {} but the output of your "
    "criterion function only contains the entries:\n{}"
)


def internal_criterion_and_derivative_template(
    x,
    *,
    task,
    direction,
    criterion,
    params,
    reparametrize_from_internal,
    convert_derivative,
    algorithm_info,
    derivative,
    criterion_and_derivative,
    numdiff_options,
    logging,
    db_kwargs,
    error_handling,
    error_penalty,
    first_criterion_evaluation,
    cache,
    cache_size,
    fixed_log_data,
):
    """Template for the internal criterion and derivative function.

    The internal criterion and derivative function only has the arguments x and task
    and algorithm_info. The other arguments will be partialed in by estimagic at some
    point. Algorithm_info and possibly even task will be partialed in by the algorithm.

    That is the reason why this function is called a template.

    Args:
        x (np.ndarray): 1d numpy array with internal parameters.
        task (str): One of "criterion", "derivative" and "criterion_and_derivative".
        direction (str): One of "maximize" or "minimize"
        criterion (callable): (partialed) user provided criterion function that takes a
            parameter dataframe as only argument and returns a scalar, an array like
            object or a dictionary. See :ref:`criterion`.
        params (pd.DataFrame): see :ref:`params`
        reparametrize_from_internal (callable): Function that takes x and returns a
            numpy array with the values of the external parameters.
        convert_derivative (callable): Function that takes the derivative of criterion
            at the external version of x and x and returns the derivative
            of the internal criterion.
        algorithm_info (dict): Dict with the following entries:
            "primary_criterion_entry": One of "value", "contributions" and
                "root_contributions" or "dict".
            "parallelizes": Bool that indicates if the algorithm calls the internal
                criterion function in parallel. If so, caching is disabled.
            "needs_scaling": bool
            "name": string
        derivative (callable, optional): (partialed) user provided function that
            calculates the first derivative of criterion. For most algorithm, this is
            the gradient of the scalar output (or "value" entry of the dict). However
            some algorithms (e.g. bhhh) require the jacobian of the "contributions"
            entry of the dict. You will get an error if you provide the wrong type of
            derivative.
        criterion_and_derivative (callable): Function that returns criterion
            and derivative as a tuple. This can be used to exploit synergies in the
            evaluation of both functions. The fist element of the tuple has to be
            exactly the same as the output of criterion. The second has to be exactly
            the same as the output of derivative.
        numdiff_options (dict): Keyword arguments for the calculation of numerical
            derivatives. See :ref:`first_derivative` for details. Note that the default
            method is changed to "forward" for speed reasons.
        logging (bool): Wether logging is used.
        db_kwargs (dict): Dictionary with entries "database", "path" and "fast_logging".
        error_handling (str): Either "raise" or "continue". Note that "continue" does
            not absolutely guarantee that no error is raised but we try to handle as
            many errors as possible in that case without aborting the optimization.
        error_penalty (dict): Dict with the entries "constant" (float) and "slope"
            (float). If the criterion or derivative raise an error and error_handling is
            "continue", return ``constant + slope * norm(params - start_params)`` where
            ``norm`` is the euclidean distance as criterion value and adjust the
            derivative accordingly. This is meant to guide the optimizer back into a
            valid region of parameter space (in direction of the start parameters).
            Note that the constant has to be high enough to ensure that the penalty is
            actually a bad function value. The default constant is 2 times the criterion
            value at the start parameters. The default slope is 0.1.
        first_criterion_evaluation (dict): Dictionary with entries "internal_params",
            "external_params", "output".
        cache (dict): Dictionary used as cache for criterion and derivative evaluations.
        cache_size (int): Number of evaluations that are kept in cache. Default 10.
        fixed_log_data (dict): Dictionary with fixed data to be saved in the database.
            Has the entries "stage" (str) and "substage" (int).

    Returns:
        float, np.ndarray or tuple: If task=="criterion" it returns the output of
            criterion which can be a float or 1d numpy array. If task=="derivative" it
            returns the first derivative of criterion, which is a numpy array.
            If task=="criterion_and_derivative" it returns both as a tuple.

    """
    if algorithm_info["primary_criterion_entry"] == "root_contributions":
        if direction == "maximize":
            msg = (
                "Optimizers that exploit a least squares structure like {} can only be "
                "used for minimization."
            )
            raise ValueError(msg.format(algorithm_info["name"]))

    x_hash = hash_array(x)
    cache_entry = cache.get(x_hash, {})

    to_dos = _determine_to_dos(task, cache_entry, derivative, criterion_and_derivative)

    caught_exceptions = []
    new_criterion, new_derivative, new_external_derivative = None, None, None
    current_params = params.copy()
    external_x = reparametrize_from_internal(x)
    current_params["value"] = external_x

    if to_dos == []:
        pass
    elif "numerical_criterion_and_derivative" in to_dos:

        def func(x):
            external_x = reparametrize_from_internal(x)
            p = params.copy()
            p["value"] = external_x
            return criterion(p)

        options = numdiff_options.copy()
        options["key"] = algorithm_info["primary_criterion_entry"]
        options["f0"] = cache_entry.get("criterion", None)
        options["return_func_value"] = True

        try:
            derivative_dict = first_derivative(func, x, **options)
            new_derivative = {
                algorithm_info["primary_criterion_entry"]: derivative_dict["derivative"]
            }
            new_criterion = derivative_dict["func_value"]
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as e:
            caught_exceptions.append(get_traceback())
            if "criterion" in cache_entry:
                raise Exception(DERIVATIVE_ERROR_MESSAGE) from e

    elif "criterion_and_derivative" in to_dos:
        try:
            new_criterion, new_external_derivative = criterion_and_derivative(
                current_params
            )
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as e:
            caught_exceptions.append(get_traceback())
            if "criterion" in cache_entry:
                raise Exception(DERIVATIVE_ERROR_MESSAGE) from e

    else:
        if "criterion" in to_dos:
            try:
                new_criterion = criterion(current_params)
            except (KeyboardInterrupt, SystemExit):
                raise
            except Exception as e:
                caught_exceptions.append(get_traceback())
                if "derivative" in cache_entry:
                    raise Exception(CRITERION_ERROR_MESSAGE) from e

        if "derivative" in to_dos:
            try:
                new_external_derivative = derivative(current_params)
            except (KeyboardInterrupt, SystemExit):
                raise
            except Exception as e:
                caught_exceptions.append(get_traceback())
                if "criterion" in cache_entry:
                    raise Exception(DERIVATIVE_ERROR_MESSAGE) from e

    if new_derivative is None and new_external_derivative is not None:
        if not isinstance(new_external_derivative, dict):
            new_external_derivative = {
                algorithm_info["primary_criterion_entry"]: new_external_derivative
            }

        new_derivative = {
            k: convert_derivative(v, internal_values=x)
            for k, v in new_external_derivative.items()
        }

    if caught_exceptions:
        if error_handling == "continue":
            new_criterion, new_derivative = _penalty_and_derivative(
                x, first_criterion_evaluation, error_penalty, algorithm_info
            )
            warnings.warn("\n\n".join(caught_exceptions))
        else:
            raise Exception("\n\n".join(caught_exceptions))

    if not algorithm_info["parallelizes"] and cache_size >= 1:
        _cache_new_evaluations(new_criterion, new_derivative, x_hash, cache, cache_size)

    new_criterion = _check_and_harmonize_criterion_output(
        cache_entry.get("criterion", new_criterion), algorithm_info
    )

    new_derivative = _check_and_harmonize_derivative(
        cache_entry.get("derivative", new_derivative), algorithm_info
    )

    if (new_criterion is not None or new_derivative is not None) and logging:
        _log_new_evaluations(
            new_criterion=new_criterion,
            new_derivative=new_derivative,
            external_x=external_x,
            caught_exceptions=caught_exceptions,
            db_kwargs=db_kwargs,
            fixed_log_data=fixed_log_data,
        )

    res = _get_output_for_optimizer(
        new_criterion, new_derivative, task, algorithm_info, direction
    )
    return res


def _determine_to_dos(task, cache_entry, derivative, criterion_and_derivative):
    """Determine which functions have to be evaluated at the new parameters.

    Args:
        task (str): One of "criterion", "derivative", "criterion_and_derivative"
        cache_entry (dict): Possibly empty dict.
        derivative (Callable or None): Only used to determine if a closed form
            derivative is available.
        criterion_and_derivative (callable or None): Only used to determine if this
            function is available.

    Returns:
        list: List of functions that have to be evaluated. Possible values are:
            - [] if nothing has to be done
            - ["criterion_and_derivative"]
            - ["numerical_criterion_and_derivative"]
            - ["criterion", "derivative"]
            - ["criterion"]
            - ["derivative"]

    """
    criterion_needed = "criterion" in task and "criterion" not in cache_entry
    derivative_needed = "derivative" in task and "derivative" not in cache_entry

    to_dos = []
    if criterion_and_derivative is not None and criterion_needed and derivative_needed:
        to_dos.append("criterion_and_derivative")
    elif (
        derivative is None
        and criterion_and_derivative is not None
        and derivative_needed
    ):
        to_dos.append("criterion_and_derivative")
    elif derivative is None and derivative_needed:
        to_dos.append("numerical_criterion_and_derivative")
    else:
        if derivative_needed:
            to_dos.append("derivative")
        if criterion_needed:
            to_dos.append("criterion")
    return to_dos


def _penalty_and_derivative(x, first_eval, error_penalty, algorithm_info):
    constant = error_penalty["constant"]
    slope = error_penalty["slope"]
    x0 = first_eval["internal_params"]

    primary = algorithm_info["primary_criterion_entry"]

    if primary == "value":
        penalty = _penalty_value(x, constant, slope, x0)
        derivative = _penalty_value_derivative(x, constant, slope, x0)
    elif primary == "contributions":
        dim_out = len(first_eval["output"][primary])
        penalty = _penalty_contributions(x, constant, slope, x0, dim_out)
        derivative = _penalty_contributions_derivative(x, constant, slope, x0, dim_out)
    elif primary == "root_contributions":
        dim_out = len(first_eval["output"][primary])
        penalty = _penalty_root_contributions(x, constant, slope, x0, dim_out)
        derivative = _penalty_root_contributions_derivative(
            x, constant, slope, x0, dim_out
        )
    else:
        raise ValueError()

    return penalty, derivative


def _penalty_value(x, constant, slope, x0, dim_out=None):
    return constant + slope * np.linalg.norm(x - x0)


def _penalty_contributions(x, constant, slope, x0, dim_out):
    contrib = (constant + slope * np.linalg.norm(x - x0)) / dim_out
    return np.ones(dim_out) * contrib


def _penalty_root_contributions(x, constant, slope, x0, dim_out):
    contrib = np.sqrt((constant + slope * np.linalg.norm(x - x0)) / dim_out)
    return np.ones(dim_out) * contrib


def _penalty_value_derivative(x, constant, slope, x0, dim_out=None):
    return slope * (x - x0) / np.linalg.norm(x - x0)


def _penalty_contributions_derivative(x, constant, slope, x0, dim_out):
    row = slope * (x - x0) / (dim_out * np.linalg.norm(x - x0))
    return np.full((dim_out, len(x)), row)


def _penalty_root_contributions_derivative(x, constant, slope, x0, dim_out):
    inner_deriv = slope * (x - x0) / np.linalg.norm(x - x0)
    outer_deriv = 0.5 / np.sqrt(_penalty_value(x, constant, slope, x0) * dim_out)
    row = outer_deriv * inner_deriv
    return np.full((dim_out, len(x)), row)


def _cache_new_evaluations(new_criterion, new_derivative, x_hash, cache, cache_size):
    cache_entry = cache.get(x_hash, {}).copy()
    if len(cache) >= cache_size:
        # list(dict) returns keys in insertion order: https://tinyurl.com/o464nrz
        oldest_entry = list(cache)[0]
        del cache[oldest_entry]
    if new_criterion is not None:
        cache_entry["criterion"] = new_criterion
    if new_derivative is not None:
        cache_entry["derivative"] = new_derivative
    cache[x_hash] = cache_entry


def _check_and_harmonize_criterion_output(output, algorithm_info):

    algo_name = algorithm_info.get("name", "your algorithm")
    primary = algorithm_info["primary_criterion_entry"]

    if output is not None:
        if np.isscalar(output):
            output = {"value": float(output)}

        if not isinstance(output, dict):
            raise ValueError("The output of criterion must be a scalar or dict.")

        if "contributions" not in output and "root_contributions" in output:
            output["contributions"] = output["root_contributions"] ** 2

        if "value" not in output and "contributions" in output:
            output["value"] = output["contributions"].sum()

        if primary not in output and primary != "dict":
            raise ValueError(
                NO_PRIMARY_MESSAGE.format(algo_name, primary, list(output))
            )
    return output


def _check_and_harmonize_derivative(derivative, algorithm_info):
    primary = algorithm_info["primary_criterion_entry"]

    if not isinstance(derivative, dict) and derivative is not None:
        derivative = {primary: derivative}

    if derivative is None:
        pass
    else:

        if primary not in derivative:
            raise ValueError(
                "If derivative returns a dict and you use an optimizer that works with "
                f"{primary}, the derivative dictionary also must contain {primary} "
                "as a key."
            )

        if "value" in derivative and np.atleast_2d(derivative["value"]).shape[0] != 1:
            raise ValueError("The derivative of a scalar optimizer must be a 1d array.")
        if "contributions" in derivative:
            if len(derivative["contributions"].shape) != 2:
                raise ValueError(
                    "The derivative of an optimizer that exploits a sum "
                    "structure must be a 2d array."
                )
        if "root_contributions" in derivative:
            if len(derivative["root_contributions"].shape) != 2:
                raise ValueError(
                    "The derivative of an optimizer that exploits a least squares "
                    "structure must be a 2d array."
                )

    return derivative


def _log_new_evaluations(
    new_criterion,
    new_derivative,
    external_x,
    caught_exceptions,
    db_kwargs,
    fixed_log_data,
):
    """Write the new evaluations and additional information into the database.

    Note: There are some seemingly unnecessary type conversions because sqlalchemy
    can fail silently when called with numpy dtypes instead of the equivalent python
    types.

    """
    data = {
        "params": external_x,
        "timestamp": datetime.datetime.now(),
        "valid": True,
        **fixed_log_data,
    }

    if new_derivative is not None:
        data["internal_derivative"] = new_derivative

    if caught_exceptions:
        separator = "\n" + "=" * 80 + "\n"
        data["exceptions"] = separator.join(caught_exceptions)
        data["valid"] = False

    if new_criterion is not None:
        data = {**data, **new_criterion}
        data["value"] = float(data["value"])

    name = "optimization_iterations"

    append_row(data, name, **db_kwargs)


def _get_output_for_optimizer(
    new_criterion, new_derivative, task, algorithm_info, direction
):
    primary = algorithm_info["primary_criterion_entry"]

    if "criterion" in task:
        if primary != "dict":
            crit = new_criterion[primary]
            crit = crit if np.isscalar(crit) else np.array(crit)
            crit = crit if direction == "minimize" else -crit
        else:
            crit = new_criterion
            if direction == "maximize":
                crit = switch_sign(crit)

    if "derivative" in task:
        deriv = np.array(new_derivative[primary])
        deriv = deriv if direction == "minimize" else -deriv

    if task == "criterion_and_derivative":
        res = (crit, deriv)
    elif task == "criterion":
        res = crit
    elif task == "derivative":
        res = deriv
    else:
        raise ValueError()

    return res



from estimagic.inference.ml_covs import cov_cluster_robust
from estimagic.inference.ml_covs import cov_hessian
from estimagic.inference.ml_covs import cov_jacobian
from estimagic.inference.ml_covs import cov_robust
from estimagic.inference.ml_covs import cov_strata_robust
from estimagic.inference.shared import calculate_inference_quantities
from estimagic.inference.shared import check_is_optimized_and_derivative_case
from estimagic.inference.shared import get_derivative_case
from estimagic.inference.shared import get_internal_first_derivative
from estimagic.inference.shared import transform_covariance
from estimagic.optimization.optimize import maximize
from estimagic.parameters.parameter_conversion import get_derivative_conversion_function
from estimagic.parameters.process_constraints import process_constraints
from estimagic.shared.check_option_dicts import check_numdiff_options
from estimagic.shared.check_option_dicts import check_optimization_options


def estimate_ml(
    loglike,
    params,
    optimize_options,
    *,
    constraints=None,
    logging=False,
    log_options=None,
    loglike_kwargs=None,
    derivative=None,
    derivative_kwargs=None,
    loglike_and_derivative=None,
    loglike_and_derivative_kwargs=None,
    numdiff_options=None,
    jacobian=None,
    jacobian_kwargs=None,
    hessian=False,
    hessian_kwargs=None,
    ci_level=0.95,
    n_samples=10_000,
    bounds_handling="raise",
    design_info=None,
):
    """Do a maximum likelihood (ml) estimation.

    This is a high level interface of our lower level functions for maximization,
    numerical differentiation and inference. It does the full workflow for maximum
    likelihood estimation with just one function call.

    While we have good defaults, you can still configure each aspect of each step
    via the optional arguments of this function. If you find it easier to do the
    "difficult" steps (mainly maximization and calculating numerical derivatives
    of a potentially noisy function) separately, you can do so and just provide those
    results as ``params``, ``jacobian`` and ``hessian``.

    The docstring is aspirational and not all options are supported yet.

    Args:
        loglike (callable): Likelihood function that takes a params DataFrame (and
            potentially other keyword arguments) and returns a dictionary that has at
            least the entries "value" (a scalar float) and "contributions" (a 1d numpy
            array or pandas Series) with the log likelihood contribution per individual.
        params (pd.DataFrame): DataFrame where the "value" column contains the
            estimated or start parameters of a likelihood model. See :ref:`params` for
            details. If the supplied parameters are estimated parameters, set
            optimize_options to False.
        optimize_options (dict or False): Keyword arguments that govern the numerical
            optimization. Valid entries are all arguments of
            :func:`~estimagic.optimization.optimize.minimize` except for criterion,
            derivative, criterion_and_derivative and params. If you pass False as
            optimize_options you signal that ``params`` are already the optimal
            parameters and no numerical optimization is needed.
        constraints (list): List with constraint dictionaries.
            See .. _link: ../../docs/source/how_to_guides/how_to_use_constraints.ipynb
        logging (pathlib.Path, str or False): Path to sqlite3 file (which typically has
            the file extension ``.db``. If the file does not exist, it will be created.
            The dashboard can only be used when logging is used.
        log_options (dict): Additional keyword arguments to configure the logging.
            - "fast_logging": A boolean that determines if "unsafe" settings are used
            to speed up write processes to the database. This should only be used for
            very short running criterion functions where the main purpose of the log
            is a real-time dashboard and it would not be catastrophic to get a
            corrupted database in case of a sudden system shutdown. If one evaluation
            of the criterion function (and gradient if applicable) takes more than
            100 ms, the logging overhead is negligible.
            - "if_table_exists": (str) One of "extend", "replace", "raise". What to
            do if the tables we want to write to already exist. Default "extend".
            - "if_database_exists": (str): One of "extend", "replace", "raise". What to
            do if the database we want to write to already exists. Default "extend".
        loglike_kwargs (dict): Additional keyword arguments for loglike.
        derivative (callable): Function takes params and potentially other keyword
            arguments and calculates the first derivative of loglike. It can either
            return a numpy array or pandas Series/DataFrame with the derivative or
            a dictionary with derivatives of each output of loglike. If loglike
            returns a dict but derivative does not, it is your responsibility to
            make sure that the correct derivative for the numerical optimizers you are
            using is returned.
        derivative_kwargs (dict): Additional keyword arguments for loglike.
        loglike_and_derivative (callable): Return a tuple consisting of the result
            of loglike and the result of derivative. Only use this if you can exploit
            synergies in the calculation of loglike and derivative.
        loglike_and_derivative_kwargs (dict): Additional keyword arguments for
            loglike_and_derivative.
        numdiff_options (dict): Keyword arguments for the calculation of numerical
            derivatives for the calculation of standard errors. See
            :ref:`first_derivative` for details.
        jacobian (callable or pandas.DataFrame or False): A function that takes
            ``params`` and potentially other keyword arguments and returns the jacobian
            of loglike["contributions"] with respect to the params. Alternatively, you
            can pass a pandas.DataFrame with the Jacobian at the optimal parameters.
            This is only possible if you pass ``optimize_options=False``. Note that you
            only need to pass a Jacobian function if you have a closed form Jacobian but
            decided not to return it as part of ``derivative`` (e.g. because you use
            a scalar optimizer and can calculate a gradient in a way that is faster
            than calculating and summing the Jacobian). If you pass None, a numerical
            Jacobian will be calculated. If you pass ``False``, you signal that no
            Jacobian should be calculated. Thus, no result that requires the Jacobian
            will be calculated.
        jacobian_kwargs (dict): Additional keyword arguments for the Jacobian function.
        hessian (callable or pd.DataFrame): A function that takes
            ``params`` and potentially other keyword arguments and returns the Hessian
            of loglike["value"] with respect to the params. Alternatively, you
            can pass a pandas.DataFrame with the Hessian at the optimal parameters.
            This is only possible if you pass ``optimize_options=False``. If you pass
            None, a numerical Hessian will be calculated. If you pass ``False``, you
            signal that no Hessian should be calculated. Thus, no result that requires
            the Hessian will be calculated.
        hessian_kwargs (dict): Additional keyword arguments for the Hessian function.
        ci_level (float): Confidence level for the calculation of confidence intervals.
            The default is 0.95.
        n_samples (int): Number of samples used to transform the covariance matrix of
            the internal parameter vector into the covariance matrix of the external
            parameters. For background information about internal and external params
            see :ref:`implementation_of_constraints`. This is only used if you have
            specified constraints.
        bounds_handling (str): One of "clip", "raise", "ignore". Determines how bounds
            are handled. If "clip", confidence intervals are clipped at the bounds.
            Standard errors are only adjusted if a sampling step is necessary due to
            additional constraints. If "raise" and any lower or upper bound is binding,
            we raise an Error. If "ignore", boundary problems are simply ignored.
        design_info (pandas.DataFrame): DataFrame with one row per observation that
            contains some or all of the variables "psu" (primary sampling unit),
            "stratum" and "fpc" (finite population corrector). See
            :ref:`robust_likelihood_inference` for details.

    Returns:
        dict: The estimated parameters, standard errors and covariance matrix of the
            parameters.

    """
    # ==================================================================================
    # Check and process inputs
    # ==================================================================================
    is_optimized = optimize_options is False

    if not is_optimized:
        check_optimization_options(
            optimize_options,
            usage="estimate_ml",
            algorithm_mandatory=True,
        )

    jac_case = get_derivative_case(jacobian)
    hess_case = get_derivative_case(hessian)

    check_is_optimized_and_derivative_case(is_optimized, jac_case)
    check_is_optimized_and_derivative_case(is_optimized, hess_case)

    cov_cases = _get_cov_cases(jac_case, hess_case, design_info)

    check_numdiff_options(numdiff_options, "estimate_ml")
    numdiff_options = {} if numdiff_options in (None, False) else numdiff_options

    constraints = [] if constraints is None else constraints

    processed_constraints, _ = process_constraints(constraints, params)

    # ==================================================================================
    # Calculate estimates via maximization (if necessary)
    # ==================================================================================

    if is_optimized:
        estimates = params
    else:
        opt_res = maximize(
            criterion=loglike,
            criterion_kwargs=loglike_kwargs,
            params=params,
            constraints=constraints,
            derivative=derivative,
            derivative_kwargs=derivative_kwargs,
            criterion_and_derivative=loglike_and_derivative,
            criterion_and_derivative_kwargs=loglike_and_derivative_kwargs,
            logging=logging,
            log_options=log_options,
            **optimize_options,
        )
        estimates = opt_res["solution_params"]

    # ==================================================================================
    # Calculate internal jacobian
    # ==================================================================================

    deriv_to_internal = get_derivative_conversion_function(
        params=params, constraints=constraints
    )

    if jac_case == "pre-calculated":
        int_jac = deriv_to_internal(jacobian)
    elif jac_case == "closed-form":
        jacobian_kwargs = {} if jacobian_kwargs is None else jacobian_kwargs
        _jac = jacobian(estimates, **jacobian_kwargs)
        int_jac = deriv_to_internal(_jac)
    # switch to "numerical" even if jac_case == "skip" because jac is required for ml.
    elif jac_case == "numerical":
        options = numdiff_options.copy()
        options["key"] = "contributions"
        deriv_res = get_internal_first_derivative(
            func=loglike,
            params=estimates,
            constraints=constraints,
            func_kwargs=loglike_kwargs,
            numdiff_options=options,
        )
        int_jac = deriv_res["derivative"]
        jac_numdiff_info = {k: v for k, v in deriv_res.items() if k != "derivative"}
    else:
        int_jac = None

    # ==================================================================================
    # Calculate internal Hessian (most of this is not yet implemented)
    # ==================================================================================

    if hess_case == "skip":
        int_hess = None
    elif hess_case == "numerical":
        raise NotImplementedError("Numerical Hessian calculation is not yet supported.")
        hess_numdiff_info = {}
    elif hess_case in ("closed-form", "pre-calculated") and constraints:
        raise NotImplementedError(
            "Closed-form or pre-calculated Hessians are not yet compatible with "
            "constraints."
        )
    else:
        int_hess = hessian(estimates, **hessian_kwargs)

    # ==================================================================================
    # Calculate all available internal cov types
    # ==================================================================================

    int_covs = {}
    if "jacobian" in cov_cases:
        int_covs["cov_jacobian"] = cov_jacobian(int_jac)
    if "hessian" in cov_cases:
        int_covs["cov_hessian"] = cov_hessian(int_hess)
    if "robust" in cov_cases:
        int_covs["cov_robust"] = cov_robust(jac=int_jac, hess=int_hess)
    if "cluster_robust" in cov_cases:
        int_covs["cov_cluster_robust"] = cov_cluster_robust(
            jac=int_jac, hess=int_hess, design_info=design_info
        )
    if "strata_robust" in cov_cases:
        int_covs["cov_strata_robust"] = cov_strata_robust(
            jac=int_jac, hess=int_hess, design_info=design_info
        )

    # ==================================================================================
    # Calculate all available external covs and summaries
    # ==================================================================================

    covs = {}
    summaries = {}
    for case in cov_cases:
        cov = transform_covariance(
            params=estimates,
            internal_cov=int_covs[f"cov_{case}"],
            constraints=constraints,
            n_samples=n_samples,
            bounds_handling=bounds_handling,
        )
        summary = calculate_inference_quantities(
            params=estimates,
            free_cov=cov,
            ci_level=ci_level,
        )

        covs[f"cov_{case}"] = cov
        summaries[f"summary_{case}"] = summary

    # ==================================================================================
    # Calculate external jac and hess (if no transforming constraints)
    # ==================================================================================

    if not processed_constraints:
        ext_jac = int_jac
        ext_hess = int_hess
    else:
        ext_jac = "No external Jacobian defined due to constraints."
        ext_hess = "No external Hessian defined due to constraints."

    # ==================================================================================
    # Construct output
    # ==================================================================================

    out = {
        **summaries,
        **covs,
        "jacobian": ext_jac,
        "hessian": ext_hess,
    }

    if not is_optimized:
        out["optimize_res"] = opt_res

    if jac_case == "numerical":
        out["jacobian_numdiff_info"] = jac_numdiff_info

    if hess_case == "numerical":
        out["hessian_numdiff_info"] = hess_numdiff_info

    return out


def _get_cov_cases(jac_case, hess_case, design_info):
    if jac_case == "skip" and hess_case == "skip":
        raise ValueError("Jacobian and Hessian cannot both be False.")
    elif jac_case == "skip" and hess_case != "skip":
        cases = ["hessian"]
    elif hess_case == "skip" and jac_case != "skip":
        cases = ["jacobian"]
    else:
        cases = ["jacobian", "hessian", "robust"]
        if design_info is not None:
            if "psu" in design_info:
                cases.append("cluster_robust")
            if {"strata", "psu", "fpc"}.issubset(design_info):
                cases.append("strata_robust")

    return cases



"""Process the user provided pc for use during the optimization.

The main purpose of this module is to convert the user provided constraints into
inputs for fast reparametrization functions. In the process, the constraints are
checked and consolidated. Consolidation means that redundant constraints are dropped
and other constraints are collected in meaningful bundles.

To improve readability, the actual code for checking and consolidation are in separate
modules.

Calls to functions doing checking are scattered across the module.
This is in order to perform each check as soon as it becomes possible, which allows
errors to be raised at a point where constraints still look similar to
what users wrote. However, some checks can only be done
after consolidation.

The challenge in making this module readable is that after each function is applied
the list of constraints and the params dataframe will be slightly different, but it
is not possible to reflect all of the changes in meaningful names because thy would
get way too long. We chose the following conventions:

As soon as a list of constraints or a params DataFrame is different from what the
user provided they are called pc (for processed constraints) and pp (for processed
params) respectively. If all constraints of a certain type, say linear, are collected
this collection is called pc_linear.

If only few columns of processed params are used in a function, it is better to
pass them as Series, to make the flow of information more explicit.

"""
import warnings

import numpy as np
import pandas as pd
from estimagic.parameters.check_constraints import check_constraints_are_satisfied
from estimagic.parameters.check_constraints import check_fixes_and_bounds
from estimagic.parameters.check_constraints import check_for_incompatible_overlaps
from estimagic.parameters.check_constraints import check_types
from estimagic.parameters.consolidate_constraints import consolidate_constraints
from estimagic.parameters.kernel_transformations import scale_to_internal
from estimagic.parameters.parameter_preprocessing import add_default_bounds_to_params
from estimagic.utilities import number_of_triangular_elements_to_dimension


def process_constraints(constraints, params, scaling_factor=None, scaling_offset=None):
    """Process, consolidate and check constraints.

    Args:
        constraints (list): List of dictionaries where each dictionary is a constraint.
        params (pd.DataFrame): see :ref:`params`.
        scaling_factor (np.ndarray or None): If None, no scaling factor is used.
        scaling_offset (np.ndarray or None): If None, no scaling offset is used.

    Returns:

        pc (list): A processed version of those constraints
            that entail actual transformations and not just fixing parameters.
        pp (pd.DataFrame): Processed params. A copy of params with additional columns:
            - _internal_lower:
              Lower bounds for the internal parameter vector. Those are derived from
              the original lower bounds and additional bounds implied by other
              constraints.
            - _internal_upper: As _internal_lower but for upper bounds.
            - _internal_free: Boolean column that is true for those parameters over
              which the optimizer will actually optimize.
            - _pre_replacements: The j_th element indicates the position of the internal
              parameter that has to be copied into the j_th position of the external
              parameter vector when reparametrizing from_internal, before any
              transformations are applied. Negative if no element has to be copied.
            - _post_replacements: As pre_replacements, but applied after the
              transformations are done.
            - _internal_fixed_value: Contains transformed versions of the fixed values
              that will become equal to the external fixed values after the
              kernel transformations are applied.
            - _is_fixed_to_value: True for parameters that are fixed to a value
            - _is_fixed_to_other: True for parameters that are fixed to another
              parameter

    """
    params = add_default_bounds_to_params(params)
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message="indexing past lexsort depth may impact performance."
        )
        params = params.copy()
        check_types(constraints)
        # selectors have to be processed before anything else happens to the params
        pc = _process_selectors(constraints, params)

        pc = _replace_pairwise_equality_by_equality(pc)
        pc = _process_linear_weights(pc, params)
        check_constraints_are_satisfied(pc, params)
        pc = _replace_increasing_and_decreasing_by_linear(pc)
        pc = _process_linear_weights(pc, params)
        pc, pp = consolidate_constraints(pc, params)
        check_for_incompatible_overlaps(pp, pc)
        check_fixes_and_bounds(pp, pc)

        int_lower, int_upper = _create_unscaled_internal_bounds(
            pp.lower_bound, pp.upper_bound, pc
        )
        pp["_internal_lower"] = int_lower
        pp["_internal_upper"] = int_upper
        pp["_internal_free"] = _create_internal_free(
            pp._is_fixed_to_value, pp._is_fixed_to_other, pc
        )

        for col in ["_internal_lower", "_internal_upper"]:
            pp[col] = _scale_bound_to_internal(
                pp[col],
                pp._internal_free,
                scaling_factor=scaling_factor,
                scaling_offset=scaling_offset,
            )
        pp["_pre_replacements"] = _create_pre_replacements(pp._internal_free)
        pp["_internal_fixed_value"] = _create_internal_fixed_value(pp._fixed_value, pc)

        return pc, pp


def _process_selectors(constraints, params):
    """Convert the query and loc field of the constraint into position based indices.

    Args:
        constraints (list): List of dictionaries where each dictionary is a constraint.
        params (pd.DataFrame): see :ref:`params`.

    Returns:
        pc (list): The resulting constraint dictionaries contain a new entry
            called 'index' that consists of the positions of the selected parameters.
            If the selected parameters are consecutive entries, the value corresponding
            to 'index' is a list of positions.

    """
    pc = []

    for constr in constraints:
        new_constr = constr.copy()

        if constr["type"] != "pairwise_equality":
            locs = [constr["loc"]] if "loc" in constr else []
            queries = [constr["query"]] if "query" in constr else []
        else:
            locs = new_constr.pop("locs", [])
            queries = new_constr.pop("queries", [])

        positions = pd.Series(data=np.arange(len(params)), index=params.index)

        indices = []
        for loc in locs:
            index = positions.loc[loc].astype(int).tolist()
            index = [index] if not isinstance(index, list) else index
            assert len(set(index)) == len(index), "Duplicates in loc are not allowed."
            indices.append(index)
        for query in queries:
            loc = params.query(query).index
            index = positions.loc[loc].astype(int).tolist()
            index = [index] if not isinstance(index, list) else index
            indices.append(index)

        if constr["type"] == "pairwise_equality":
            assert (
                len(indices) >= 2
            ), "Select at least 2 sets of parameters for pairwise equality constraint!"
            length = len(indices[0])
            for index in indices:
                assert len(index) == length, (
                    "All sets of parameters in pairwise_equality pc must have "
                    "the same length."
                )
        else:
            assert (
                len(indices) == 1
            ), "Either loc or query can be in constraint but not both."

        n_selected = len(indices[0])
        if n_selected >= 1:
            if constr["type"] == "pairwise_equality":
                new_constr["indices"] = indices
            else:
                new_constr["index"] = indices[0]
            pc.append(new_constr)

    return pc


def _replace_pairwise_equality_by_equality(pc):
    """Rewrite pairwise equality constraints to equality constraints.

    Args:
        pc (list): List of dictionaries where each dictionary is a constraint.
            It is assumed that the selectors in constraints were already processed.

    Returns:
        pc (list): List of processed constraints.

    """
    pairwise_constraints = [c for c in pc if c["type"] == "pairwise_equality"]
    pc = [c for c in pc if c["type"] != "pairwise_equality"]
    for constr in pairwise_constraints:
        equality_constraints = []
        for elements in zip(*constr["indices"]):
            equality_constraints.append({"index": list(elements), "type": "equality"})
        pc += equality_constraints

    return pc


def _process_linear_weights(pc, params):
    """Harmonize the weights of linear constraints.

    Args:
        pc (list): Constraints where the selector have already been processed.
        params (pd.DataFrame): see :ref:`params`.

    Returns:
        processed (list): Constraints where all weights are Series.

    """
    processed = []
    for constr in pc:
        if constr["type"] == "linear":
            raw_weights = constr["weights"]
            params_subset = params.iloc[constr["index"]]

            msg = f"Weights must be same length as selected parameters: {params_subset}"
            if isinstance(raw_weights, pd.Series):
                weights = raw_weights.loc[params_subset.index].to_numpy()
            elif isinstance(raw_weights, (np.ndarray, list, tuple)):
                if len(raw_weights) != len(params_subset):
                    raise ValueError(msg)
                weights = np.asarray(raw_weights)
            elif isinstance(raw_weights, (float, int)):
                weights = np.full(len(params_subset), float(raw_weights))
            else:
                raise TypeError(f"Invalid type for linear weights {type(raw_weights)}.")

            new_constr = constr.copy()
            weights_sr = pd.Series(weights, index=params_subset.index)
            new_constr["weights"] = weights_sr
            processed.append(new_constr)
        else:
            processed.append(constr)

    return processed


def _replace_increasing_and_decreasing_by_linear(pc):
    """Write increasing and decreasing constraints as linear constraints.

    Args:
        pc (list): Constraints where the selectors have already been processed.

    Returns:
        processed (list)

    """
    increasing_ilocs, other_constraints = [], []

    for constr in pc:
        if constr["type"] == "increasing":
            increasing_ilocs.append(constr["index"])
        elif constr["type"] == "decreasing":
            increasing_ilocs.append(constr["index"][::-1])
        else:
            other_constraints.append(constr)

    linear_constraints = []
    for iloc in increasing_ilocs:
        for smaller, larger in zip(iloc, iloc[1:]):
            linear_constr = {
                "index": [smaller, larger],
                "type": "linear",
                "weights": np.array([-1, 1]),
                "lower_bound": 0,
            }
            linear_constraints.append(linear_constr)

    processed = linear_constraints + other_constraints
    return processed


def _create_unscaled_internal_bounds(lower, upper, pc):
    """Create columns with bounds for the internal parameter vector.

    The columns have the length of the external params and will be reduced later.

    Args:
        lower (pd.Series): Processed and consolidated external lower bounds.
        upper (pd.Series): Processed and consolidated external upper bounds.
        pc (pd.DataFrame): Processed and consolidated constraints.

    Returns:
        int_lower (pd.Series): Lower bound of internal parameters.
        int_upper (pd.Series): Upper bound of internal parameters.

    """
    int_lower, int_upper = lower.copy(), upper.copy()

    for constr in pc:
        if constr["type"] in ["covariance", "sdcorr"]:
            # Note that the diagonal positions are the same for covariance and sdcorr
            # because the internal params contains the Cholesky factor of the implied
            # covariance matrix in both cases.
            dim = number_of_triangular_elements_to_dimension(len(constr["index"]))
            diag_positions = [0] + np.cumsum(range(2, dim + 1)).tolist()
            diag_indices = np.array(constr["index"])[diag_positions].tolist()
            bd = constr.get("bounds_distance", 0)
            bd = np.sqrt(bd) if constr["type"] == "covariance" else bd
            int_lower.iloc[diag_indices] = np.maximum(int_lower.iloc[diag_indices], bd)
        elif constr["type"] == "probability":
            int_lower.iloc[constr["index"]] = 0
        elif constr["type"] == "linear":
            int_lower.iloc[constr["index"]] = -np.inf
            int_upper.iloc[constr["index"]] = np.inf
            int_lower.update(constr["right_hand_side"]["lower_bound"])
            int_upper.update(constr["right_hand_side"]["upper_bound"])
        else:
            raise TypeError("Invalid constraint type {}".format(constr["type"]))

    return int_lower, int_upper


def _create_internal_free(is_fixed_to_value, is_fixed_to_other, pc):
    """Boolean Series that is true for parameters over which the optimizer optimizes.

    Args:
        is_fixed_to_value (pd.Series): The _is_fixed_to_value column of pp.
        is_fixed_to_other (pd.Series): The _is_fixed_to_other column of pp.

    Returns:
        int_free (pd.Series)
    """
    int_fixed = is_fixed_to_value | is_fixed_to_other

    for constr in pc:
        if constr["type"] == "probability":
            int_fixed.iloc[constr["index"][-1]] = True
        elif constr["type"] == "linear":
            int_fixed.iloc[constr["index"]] = False
            int_fixed.update(constr["right_hand_side"]["value"].notnull())
            # dtype gets messed up by update
            int_fixed = int_fixed.astype(bool)

    int_free = ~int_fixed

    return int_free


def _create_pre_replacements(internal_free):
    """Series with internal position of parameters.

    The j_th element indicates the position of the internal parameter that has to be
    copied into the j_th position of the external parameter vector when reparametrizing
    from_internal, before any transformations are applied. Negative if no element has
    to be copied.

    This will be used to copy the free internal parameters into a parameter vector
    that has the same length as all params.

    Args:
        internal_free (pd.Series): The _internal_free column of the processed params.

    """
    pre_replacements = (
        internal_free.replace(False, np.nan).cumsum().subtract(1).fillna(-1).astype(int)
    )

    return pre_replacements


def _create_internal_fixed_value(fixed_value, pc):
    """Pandas Series containing the values to which internal parameters are fixe.

    This contains additional fixes used to enforce other constraints and (potentially
    transformed) user specified fixed values.

    Args:
        fixed_value (pd.Series): The (external) _fixed_value column of pp.
        pc (list): Processed and consolidated params.

    """
    int_fix = fixed_value.copy()
    for constr in pc:
        if constr["type"] == "probability":
            int_fix.iloc[constr["index"][-1]] = 1
        elif constr["type"] in ["covariance", "sdcorr"]:
            int_fix.iloc[constr["index"][0]] = np.sqrt(int_fix.iloc[constr["index"][0]])
        elif constr["type"] == "linear":
            int_fix.iloc[constr["index"]] = np.nan
            int_fix.update(constr["right_hand_side"]["value"])

    return int_fix


def _scale_bound_to_internal(bound_sr, internal_free, scaling_factor, scaling_offset):
    sr = bound_sr.copy(deep=True)
    free_bounds = bound_sr[internal_free].to_numpy()

    scaled = scale_to_internal(free_bounds, scaling_factor, scaling_offset)

    sr[internal_free] = scaled
    return sr



"""Functions to calculate and apply scaling factors and offsets.

All scaling operations are applied to internal parameter vectors, i.e.
after the reparametrizations for constraints have been applied.

The order of applying the operations is the following:

to internal: (external - scaling_offset) / scaling_factor
from internal: internal * scaling_factor + scaling_offset

"""
import numpy as np
from estimagic.differentiation.derivatives import first_derivative
from estimagic.parameters.parameter_conversion import get_internal_bounds
from estimagic.parameters.parameter_conversion import get_reparametrize_functions


def calculate_scaling_factor_and_offset(
    params,
    constraints,
    criterion,
    method="start_values",
    clipping_value=0.1,
    magnitude=1,
    numdiff_options=None,
    processed_params=None,
    processed_constraints=None,
):
    numdiff_options = {} if numdiff_options is None else numdiff_options
    to_internal, from_internal = get_reparametrize_functions(
        params=params,
        constraints=constraints,
        processed_params=processed_params,
        processed_constraints=processed_constraints,
    )

    x = to_internal(params["value"].to_numpy())

    if method in ("bounds", "gradient"):
        lower_bounds, upper_bounds = get_internal_bounds(
            params, constraints, processed_params=processed_params
        )

    if method == "start_values":
        raw_factor = np.clip(np.abs(x), clipping_value, np.inf)
        scaling_offset = None
    elif method == "bounds":
        raw_factor = upper_bounds - lower_bounds
        scaling_offset = lower_bounds
    elif method == "gradient":
        default_numdiff_options = {
            "scaling_factor": 100,
            "lower_bounds": lower_bounds,
            "upper_bounds": upper_bounds,
            "error_handling": "raise",
        }

        numdiff_options = {**default_numdiff_options, **numdiff_options}

        def func(x):
            p = params.copy(deep=True)
            p["value"] = from_internal(x)
            crit = criterion(p)
            if isinstance(crit, dict):
                crit = crit["value"]
            return crit

        gradient = first_derivative(func, x, **numdiff_options)["derivative"]

        raw_factor = np.clip(np.abs(gradient), clipping_value, np.inf)
        scaling_offset = None

    scaling_factor = raw_factor / magnitude

    return scaling_factor, scaling_offset



"""High level functions to convert between parameter vectors.

High level means:

- Functions are save to use directly with user input (if applicable)
- Defaults are filled automatically (if applicable)
- Robust checks and error handling

"""
import functools

import numpy as np
from estimagic.parameters.kernel_transformations import scale_from_internal
from estimagic.parameters.kernel_transformations import scale_to_internal
from estimagic.parameters.parameter_preprocessing import add_default_bounds_to_params
from estimagic.parameters.parameter_preprocessing import check_params_are_valid
from estimagic.parameters.process_constraints import process_constraints
from estimagic.parameters.reparametrize import convert_external_derivative_to_internal
from estimagic.parameters.reparametrize import post_replace_jacobian
from estimagic.parameters.reparametrize import pre_replace_jacobian
from estimagic.parameters.reparametrize import reparametrize_from_internal
from estimagic.parameters.reparametrize import reparametrize_to_internal


def get_reparametrize_functions(
    params,
    constraints,
    scaling_factor=None,
    scaling_offset=None,
    processed_params=None,
    processed_constraints=None,
):
    """Construct functions to map between internal and external parameters.

    All required information is partialed into the functions.

    Args:
        params (pandas.DataFrame): See :ref:`params`.
        constraints (list): List of constraint dictionaries.
        scaling_factor (np.ndarray or None): If None, no scaling factor is used.
        scaling_offset (np.ndarray or None): If None, no scaling offset is used.
        processed_params (pandas.DataFrame): Processed parameters.
        processed_constraints (list): Processed constraints.

    Returns:
        func: Function that maps an external parameter vector to an internal one
        func: Function that maps an internal parameter vector to an external one

    """
    if constraints in [None, []]:
        partialed_to_internal = functools.partial(
            no_constraint_to_internal,
            scaling_factor=scaling_factor,
            scaling_offset=scaling_offset,
        )
        partialed_from_internal = functools.partial(
            no_constraint_from_internal,
            params=params,
            scaling_factor=scaling_factor,
            scaling_offset=scaling_offset,
        )
    else:
        if processed_params is None or processed_constraints is None:
            params = add_default_bounds_to_params(params)
            check_params_are_valid(params)

            processed_constraints, processed_params = process_constraints(
                constraints=constraints,
                params=params,
                scaling_factor=scaling_factor,
                scaling_offset=scaling_offset,
            )

        # get partialed reparametrize from internal
        pre_replacements = processed_params["_pre_replacements"].to_numpy()
        post_replacements = processed_params["_post_replacements"].to_numpy()
        fixed_values = processed_params["_internal_fixed_value"].to_numpy()

        # get partialed reparametrize to internal
        internal_free = processed_params["_internal_free"].to_numpy()

        partialed_to_internal = functools.partial(
            reparametrize_to_internal,
            internal_free=internal_free,
            processed_constraints=processed_constraints,
            scaling_factor=scaling_factor,
            scaling_offset=scaling_offset,
        )

        partialed_from_internal = functools.partial(
            reparametrize_from_internal,
            fixed_values=fixed_values,
            pre_replacements=pre_replacements,
            processed_constraints=processed_constraints,
            post_replacements=post_replacements,
            params=params,
            scaling_factor=scaling_factor,
            scaling_offset=scaling_offset,
        )

    return partialed_to_internal, partialed_from_internal


def get_derivative_conversion_function(
    params,
    constraints,
    scaling_factor=None,
    scaling_offset=None,
    processed_params=None,
    processed_constraints=None,
):
    """Construct functions to map between internal and external derivatives.

    All required information is partialed into the functions.

    Args:
        params (pandas.DataFrame): See :ref:`params`.
        constraints (list): List of constraint dictionaries.
        scaling_factor (np.ndarray or None): If None, no scaling factor is used.
        scaling_offset (np.ndarray or None): If None, no scaling offset is used.
        processed_params (pandas.DataFrame): Processed parameters.
        processed_constraints (list): Processed constraints.


    Returns:
        func: Function that converts an external derivative to an internal one

    """
    if constraints in [None, []]:
        convert_derivative = functools.partial(
            no_constraint_derivative_to_internal,
            scaling_factor=scaling_factor,
        )

    else:
        if processed_params is None or processed_constraints is None:
            params = add_default_bounds_to_params(params)
            check_params_are_valid(params)
            processed_constraints, processed_params = process_constraints(
                constraints=constraints,
                params=params,
                scaling_factor=scaling_factor,
                scaling_offset=scaling_offset,
            )

        pre_replacements = processed_params["_pre_replacements"].to_numpy()
        post_replacements = processed_params["_post_replacements"].to_numpy()
        fixed_values = processed_params["_internal_fixed_value"].to_numpy()

        dim_internal = int(processed_params["_internal_free"].sum())

        pre_replace_jac = pre_replace_jacobian(
            pre_replacements=pre_replacements, dim_in=dim_internal
        )
        post_replace_jac = post_replace_jacobian(post_replacements=post_replacements)

        convert_derivative = functools.partial(
            convert_external_derivative_to_internal,
            fixed_values=fixed_values,
            pre_replacements=pre_replacements,
            processed_constraints=processed_constraints,
            pre_replace_jac=pre_replace_jac,
            post_replace_jac=post_replace_jac,
            scaling_factor=scaling_factor,
            scaling_offset=scaling_offset,
        )

    return convert_derivative


def get_internal_bounds(
    params, constraints, scaling_factor=None, scaling_offset=None, processed_params=None
):
    if constraints in [None, []]:
        params = add_default_bounds_to_params(params)
        lower_bounds = params["lower_bound"]
        upper_bounds = params["upper_bound"]
    else:
        if processed_params is None:
            params = add_default_bounds_to_params(params)
            check_params_are_valid(params)

            _, processed_params = process_constraints(
                constraints=constraints,
                params=params,
                scaling_factor=scaling_factor,
                scaling_offset=scaling_offset,
            )

        free = processed_params[processed_params["_internal_free"]]
        lower_bounds = free["_internal_lower"].to_numpy()
        upper_bounds = free["_internal_upper"].to_numpy()
    return lower_bounds, upper_bounds


def no_constraint_to_internal(external, scaling_factor, scaling_offset):
    if isinstance(external, np.ndarray):
        unscaled = external
    else:
        unscaled = external["value"].to_numpy()

    out = scale_to_internal(unscaled, scaling_factor, scaling_offset)

    return out


def no_constraint_from_internal(
    internal, params, scaling_factor, scaling_offset, return_numpy=True
):

    scaled = scale_from_internal(internal, scaling_factor, scaling_offset)

    if return_numpy:
        out = scaled
    else:
        out = params.copy()
        out["value"] = internal
    return out


def no_constraint_derivative_to_internal(
    external_derivative, internal_values, scaling_factor
):  # noqa: U100
    if scaling_factor is not None:
        out = external_derivative @ np.diag(scaling_factor)
    else:
        out = external_derivative
    return out



import inspect

from estimagic.config import IS_CYIPOPT_INSTALLED
from estimagic.config import IS_DFOLS_INSTALLED
from estimagic.config import IS_FIDES_INSTALLED
from estimagic.config import IS_NLOPT_INSTALLED
from estimagic.config import IS_PETSC4PY_INSTALLED
from estimagic.config import IS_PYBOBYQA_INSTALLED
from estimagic.config import IS_PYGMO_INSTALLED
from estimagic.optimization import cyipopt_optimizers
from estimagic.optimization import fides_optimizers
from estimagic.optimization import nag_optimizers
from estimagic.optimization import nlopt_optimizers
from estimagic.optimization import pygmo_optimizers
from estimagic.optimization import scipy_optimizers
from estimagic.optimization import tao_optimizers
from estimagic.optimization.bhhh import bhhh
from estimagic.optimization.neldermead import neldermead_parallel
from estimagic.optimization.pounders import pounders


COLLECTED_FUNCTIONS = {
    **dict(inspect.getmembers(scipy_optimizers, inspect.isfunction)),
}

if IS_PETSC4PY_INSTALLED:
    COLLECTED_FUNCTIONS.update(
        **dict(inspect.getmembers(tao_optimizers, inspect.isfunction))
    )

if IS_NLOPT_INSTALLED:
    COLLECTED_FUNCTIONS.update(
        **dict(inspect.getmembers(nlopt_optimizers, inspect.isfunction))
    )

# drop private and helper functions
AVAILABLE_ALGORITHMS = {
    "bhhh": bhhh,
    "neldermead_parallel": neldermead_parallel,
    "pounders": pounders,
}

PUBLIC_HELPERS = [
    "calculate_trustregion_initial_radius",
    "get_scipy_bounds",
    "process_scipy_result",
]
for k, v in COLLECTED_FUNCTIONS.items():
    if not k.startswith("_") and k not in PUBLIC_HELPERS:
        AVAILABLE_ALGORITHMS[k] = v

if IS_PYBOBYQA_INSTALLED:
    AVAILABLE_ALGORITHMS["nag_pybobyqa"] = nag_optimizers.nag_pybobyqa

if IS_DFOLS_INSTALLED:
    AVAILABLE_ALGORITHMS["nag_dfols"] = nag_optimizers.nag_dfols

if IS_PYGMO_INSTALLED:
    _PYGMO_FUNCTIONS = dict(inspect.getmembers(pygmo_optimizers, inspect.isfunction))
    PYGMO_ALGORITHMS = {
        k: v for k, v in _PYGMO_FUNCTIONS.items() if k.startswith("pygmo_")
    }
    AVAILABLE_ALGORITHMS.update(**PYGMO_ALGORITHMS)

if IS_CYIPOPT_INSTALLED:
    AVAILABLE_ALGORITHMS["ipopt"] = cyipopt_optimizers.ipopt

if IS_FIDES_INSTALLED:
    AVAILABLE_ALGORITHMS["fides"] = fides_optimizers.fides

GLOBAL_ALGORITHMS = [
    "nlopt_direct",
    "nlopt_esch",
    "nlopt_isres",
    "nlopt_crs2_lm",
]
if IS_PYGMO_INSTALLED:
    GLOBAL_ALGORITHMS += PYGMO_ALGORITHMS.keys()



def process_internal_optimizer_result(
    res,
    direction,
    params_from_internal,
):
    """Process results of internal optimizers.

    Args:
        res (dict): Results dictionary of an internal optimizer or multistart optimizer.
        direction (str): One of "maximize" or "minimize". Used to switch sign of
            criterion evaluations.
        params_from_internal (callable): Function that converts internal parameters
            to external ones.

    """
    if isinstance(res, str):
        res = _dummy_result_from_traceback(res)
    else:
        res = _process_one_result(res, direction, params_from_internal)

        if "multistart_info" in res:
            res["multistart_info"] = _process_multistart_info(
                res["multistart_info"],
                direction,
                params_from_internal,
            )
    return res


def _process_one_result(res, direction, params_from_internal):
    res = res.copy()
    p = params_from_internal(res["solution_x"], return_numpy=False)
    res["solution_params"] = p

    if direction == "maximize" and "solution_criterion" in res:
        res["solution_criterion"] = switch_sign(res["solution_criterion"])

    # in the long run we can get some of those from the database if logging was used.
    optional_entries = [
        "solution_criterion",
        "solution_derivative",
        "solution_hessian",
        "n_criterion_evaluations",
        "n_derivative_evaluations",
        "n_iterations",
        "success",
        "reached_convergence_criterion",
        "message",
    ]

    for entry in optional_entries:
        res[entry] = res.get(entry)
    return res


def _process_multistart_info(info, direction, params_from_internal):

    starts = []
    for x in info["start_parameters"]:
        starts.append(params_from_internal(x, return_numpy=False))

    optima = []
    for res in info["local_optima"]:
        processed = _process_one_result(
            res,
            direction=direction,
            params_from_internal=params_from_internal,
        )
        optima.append(processed)

    sample = []
    for x in info["exploration_sample"]:
        sample.append(params_from_internal(x, return_numpy=False))

    if direction == "minimize":
        exploration_res = info["exploration_results"]
    else:
        exploration_res = []
        for res in info["exploration_results"]:
            exploration_res.append(switch_sign(res))

    out = {
        "start_parameters": starts,
        "local_optima": optima,
        "exploration_sample": sample,
        "exploration_results": exploration_res,
    }
    return out


def _dummy_result_from_traceback(candidate):
    if isinstance(candidate, str):
        out = {
            "solution_params": None,
            "solution_criterion": None,
            "solution_derivative": None,
            "solution_hessian": None,
            "n_criterion_evaluations": None,
            "n_derivative_evaluations": None,
            "n_iterations": None,
            "success": False,
            "reached_convergence_criterion": None,
            "message": candidate,
        }
    else:
        out = candidate
    return out


def switch_sign(critval):
    if isinstance(critval, dict):
        out = critval.copy()
        if "value" in critval:
            out["value"] = -critval["value"]
        if "contributions" in critval:
            out["contributions"] = -critval["contributions"]
    else:
        out = -critval
    return out



import numpy as np
from estimagic.parameters.tree_registry import get_registry
from pybaum import tree_just_flatten as tree_leaves
from pybaum import tree_map
from pybaum import tree_update


def get_bounds(params, lower_bounds=None, upper_bounds=None):
    """Consolidate lower/upper bounds with bounds available in params.

    Updates bounds defined in params. If no bounds are available the entry is set to
    -np.inf for the lower bound and np.inf for the upper bound. If a bound is defined in
    params and lower_bounds or upper_bounds, the bound from lower_bounds or upper_bounds
    will be used.

    Args:
        params (pytree): The parameter pytree.
        lower_bounds (pytree): Must be a subtree of params.
        upper_bounds (pytree): Must be a subtree of params.

    Returns:
        np.ndarray: Consolidated and flattened lower_bounds.
        np.ndarray: Consolidated and flattened upper_bounds.

    """
    registry = get_registry(extended=True)
    n_params = len(tree_leaves(params, registry=registry))

    # Fill leaves with np.nan. If params contains a data frame with bounds column, that
    # column is not overwritten.
    bounds_tree = tree_map(lambda leaf: np.nan, params, registry=registry)

    lower_flat = _update_bounds_and_flatten(
        bounds_tree, lower_bounds, direction="lower_bound"
    )
    upper_flat = _update_bounds_and_flatten(
        bounds_tree, upper_bounds, direction="upper_bound"
    )

    if len(lower_flat) != n_params:
        raise ValueError("lower_bounds do not match dimension of params.")
    if len(upper_flat) != n_params:
        raise ValueError("upper_bounds do not match dimension of params.")

    lower_flat[np.isnan(lower_flat)] = -np.inf
    upper_flat[np.isnan(upper_flat)] = np.inf

    return lower_flat, upper_flat


def _update_bounds_and_flatten(bounds_tree, bounds, direction):
    registry = get_registry(extended=True, data_col=direction)
    if bounds is not None:
        bounds_tree = tree_update(bounds_tree, bounds)
    bounds_flat = tree_leaves(bounds_tree, registry=registry)
    bounds_flat = np.array(bounds_flat, dtype=np.float64)
    return bounds_flat



"""Visualize and compare derivative estimates."""
import itertools

import matplotlib.pyplot as plt
import numpy as np


def derivative_plot(
    derivative_result,
):
    """Plot evaluations and derivative estimates.

    The resulting grid plot displays function evaluations and derivatives. The
    derivatives are visualized as a first-order Taylor approximation. Bands are drawn
    indicating the area in which forward and backward derivatives are located. This is
    done by filling the area between the derivative estimate with lowest and highest
    step size, respectively. Do not confuse these bands with statistical errors.

    This function does not require the params vector as plots are displayed relative to
    the point at which the derivative is calculated.

    Args:
        derivative_result (dict): The result dictionary of call to
            :func:`~estimagic.differentiation.derivatives.first_derivative` with
            return_info and return_func_value set to True.

    Returns:
        fig (matplotlib.pyplot.figure): The figure.

    """
    func_value = derivative_result["func_value"]
    func_evals = derivative_result["func_evals"]
    derivative_candidates = derivative_result["derivative_candidates"]

    # remove index from main data for plotting
    df = func_evals.reset_index()
    df = df.assign(**{"step": df.step * df.sign})
    func_evals = df.set_index(["sign", "step_number", "dim_x", "dim_f"])

    # prepare derivative data
    df_der = _select_derivative_with_minimal_error(derivative_candidates)
    df_der_method = _select_derivative_with_minimal_error(
        derivative_candidates, given_method=True
    )

    # auxiliary
    grid_points = 2  # we do not need more than 2 grid points since all lines are affine
    func_value = np.atleast_1d(func_value)
    max_steps = df.groupby("dim_x")["step"].max()
    palette = {
        "forward": "tab:green",
        "central": "tab:blue",
        "backward": "tab:orange",
        1: "green",
        -1: "orange",
    }

    # dimensions of problem. dimensions of params vector span the vertical axis while
    # dimensions of output span the horizontal axis of produced figure
    dim_x = range(df["dim_x"].max() + 1)
    dim_f = range(df["dim_f"].max() + 1)

    # plot
    width = 10 * len(dim_f)
    height = 11 * len(dim_x)

    fig, axes = plt.subplots(len(dim_x), len(dim_f), figsize=(width, height))
    axes = np.atleast_2d(axes)

    for ax, (row, col) in zip(axes.flatten(), itertools.product(dim_x, dim_f)):
        # labels and texts
        ax.set_xlabel(rf"Value relative to $x_{{0, {row}}}$", fontsize=14)
        ax.text(
            0.35,
            1.02,
            f"dim_x, dim_f = {row, col}",
            transform=ax.transAxes,
            color="grey",
            fontsize=14,
        )

        # initial values and x grid
        y0 = func_value[col]
        x_grid = np.linspace(-max_steps[row], max_steps[row], grid_points)

        # plot function evaluations scatter points
        _scatter_data = func_evals.query("dim_x == @row & dim_f == @col")
        ax.scatter(
            _scatter_data["step"],
            _scatter_data["eval"],
            color="gray",
            label="Function Evaluation",
            edgecolor="black",
        )

        # draw overall best derivative estimate
        _y = y0 + x_grid * df_der.loc[row, col]
        ax.plot(
            x_grid,
            _y,
            color="black",
            label="Best Estimate",
            zorder=2,
            linewidth=1.5,
            linestyle="dashdot",
        )

        # draw best derivative estimate given each method
        for method in ["forward", "central", "backward"]:
            _y = y0 + x_grid * df_der_method.loc[method, row, col]
            ax.plot(
                x_grid, _y, color=palette[method], label=method, zorder=1, linewidth=2
            )

        # fill area
        for sign in [1, -1]:
            _x_y = _select_eval_with_lowest_and_highest_step(func_evals, sign, row, col)
            diff = _x_y - np.array([0, y0])
            slope = diff[:, 1] / diff[:, 0]
            _y = y0 + x_grid * slope.reshape(-1, 1)
            ax.plot(x_grid, _y.T, "--", color=palette[sign], linewidth=0.5)
            ax.fill_between(x_grid, _y[0, :], _y[1, :], alpha=0.15, color=palette[sign])

    # legend
    ncol = 5 if len(dim_f) > 1 else 3
    axes[0, 0].legend(
        loc="upper center",
        bbox_to_anchor=(len(dim_f) / 2 + 0.05 * len(dim_f), 1.15),
        ncol=ncol,
        fontsize=14,
    )
    return fig


def _select_derivative_with_minimal_error(df_jac_cand, given_method=False):
    """Select derivatives with minimal error component wise.

    Args:
        df_jac_cand (pandas.DataFrame): Frame containing jacobian candidates.
        given_method (bool): Boolean indicating wether to condition on columns method
            in df_jac_cand. Default is False, which selects the overall best derivative
            estimate.

    Returns:
        df (pandas.DataFrame): The (best) derivative estimate.

    """
    given = ["method"] if given_method else []
    minimizer = df_jac_cand.groupby(given + ["dim_x", "dim_f"])["err"].idxmin()
    df = df_jac_cand.loc[minimizer]["der"]
    index_level_to_drop = list({"method", "num_term"} - set(given))
    df = df.droplevel(index_level_to_drop).copy()
    return df


def _select_eval_with_lowest_and_highest_step(df_evals, sign, dim_x, dim_f):
    """Select step and eval from data with highest and lowest step.

    Args:
        df_evals (pd.DataFrame): Frame containing func evaluations (long-format).
        sign (int): Direction of step.
        dim_x (int): Dimension of x to select.
        dim_f (int): Dimension of f to select.

    Returns:
        out (numpy.ndarray): Array of shape (2, 2). Columns correspond to step and eval,
            while rows correspond to lowest and highest step, respectively.

    """
    df = df_evals.loc[(sign, slice(None), dim_x, dim_f), ["step", "eval"]]
    df = df.dropna().sort_index()
    out = df.head(1).append(df.tail(1)).values.copy()
    return out



"""Implement algorithms by the (Numerical Algorithms Group)[https://www.nag.com/].

The following arguments are not supported as ``algo_options``:

- ``scaling_within_bounds``
- ``init.run_in_parallel``
- ``do_logging``, ``print_progress`` and all their advanced options.
  Use estimagic's database and dashboard instead to explore your criterion
  and algorithm.

"""
import warnings
from functools import partial

import numpy as np
from estimagic.config import IS_DFOLS_INSTALLED
from estimagic.config import IS_PYBOBYQA_INSTALLED
from estimagic.optimization.algo_options import CLIP_CRITERION_IF_OVERFLOWING
from estimagic.optimization.algo_options import (
    CONVERGENCE_MINIMAL_TRUSTREGION_RADIUS_TOLERANCE,
)
from estimagic.optimization.algo_options import (
    CONVERGENCE_NOISE_CORRECTED_CRITERION_TOLERANCE,
)
from estimagic.optimization.algo_options import CONVERGENCE_SLOW_PROGRESS
from estimagic.optimization.algo_options import INITIAL_DIRECTIONS
from estimagic.optimization.algo_options import INTERPOLATION_ROUNDING_ERROR
from estimagic.optimization.algo_options import RANDOM_DIRECTIONS_ORTHOGONAL
from estimagic.optimization.algo_options import RESET_OPTIONS
from estimagic.optimization.algo_options import STOPPING_MAX_CRITERION_EVALUATIONS
from estimagic.optimization.algo_options import THRESHOLD_FOR_SAFETY_STEP
from estimagic.optimization.algo_options import TRUSTREGION_EXPANSION_FACTOR_SUCCESSFUL
from estimagic.optimization.algo_options import (
    TRUSTREGION_EXPANSION_FACTOR_VERY_SUCCESSFUL,
)
from estimagic.optimization.algo_options import TRUSTREGION_FAST_START_OPTIONS
from estimagic.optimization.algo_options import TRUSTREGION_PRECONDITION_INTERPOLATION
from estimagic.optimization.algo_options import (
    TRUSTREGION_SHRINKING_FACTOR_LOWER_RADIUS,
)
from estimagic.optimization.algo_options import (
    TRUSTREGION_SHRINKING_FACTOR_NOT_SUCCESSFUL,
)
from estimagic.optimization.algo_options import (
    TRUSTREGION_SHRINKING_FACTOR_UPPER_RADIUS,
)
from estimagic.optimization.algo_options import TRUSTREGION_THRESHOLD_SUCCESSFUL
from estimagic.optimization.algo_options import TRUSTREGION_THRESHOLD_VERY_SUCCESSFUL
from estimagic.utilities import calculate_trustregion_initial_radius

if IS_PYBOBYQA_INSTALLED:
    import pybobyqa

if IS_DFOLS_INSTALLED:
    import dfols


def nag_dfols(
    criterion_and_derivative,
    x,
    lower_bounds,
    upper_bounds,
    *,
    clip_criterion_if_overflowing=CLIP_CRITERION_IF_OVERFLOWING,
    convergence_minimal_trustregion_radius_tolerance=CONVERGENCE_MINIMAL_TRUSTREGION_RADIUS_TOLERANCE,  # noqa: E501
    convergence_noise_corrected_criterion_tolerance=CONVERGENCE_NOISE_CORRECTED_CRITERION_TOLERANCE,  # noqa: E501
    convergence_scaled_criterion_tolerance=0.0,
    convergence_slow_progress=None,
    initial_directions=INITIAL_DIRECTIONS,
    interpolation_rounding_error=INTERPOLATION_ROUNDING_ERROR,
    noise_additive_level=None,
    noise_multiplicative_level=None,
    noise_n_evals_per_point=None,
    random_directions_orthogonal=RANDOM_DIRECTIONS_ORTHOGONAL,
    stopping_max_criterion_evaluations=STOPPING_MAX_CRITERION_EVALUATIONS,
    threshold_for_safety_step=THRESHOLD_FOR_SAFETY_STEP,
    trustregion_expansion_factor_successful=TRUSTREGION_EXPANSION_FACTOR_SUCCESSFUL,
    trustregion_expansion_factor_very_successful=TRUSTREGION_EXPANSION_FACTOR_VERY_SUCCESSFUL,  # noqa: E501
    trustregion_fast_start_options=None,
    trustregion_initial_radius=None,
    trustregion_method_to_replace_extra_points="geometry_improving",
    trustregion_n_extra_points_to_replace_successful=0,
    trustregion_n_interpolation_points=None,
    trustregion_precondition_interpolation=TRUSTREGION_PRECONDITION_INTERPOLATION,
    trustregion_reset_options=None,
    trustregion_shrinking_factor_not_successful=TRUSTREGION_SHRINKING_FACTOR_NOT_SUCCESSFUL,  # noqa: E501
    trustregion_shrinking_factor_lower_radius=TRUSTREGION_SHRINKING_FACTOR_LOWER_RADIUS,
    trustregion_shrinking_factor_upper_radius=TRUSTREGION_SHRINKING_FACTOR_UPPER_RADIUS,
    trustregion_threshold_successful=TRUSTREGION_THRESHOLD_SUCCESSFUL,
    trustregion_threshold_very_successful=TRUSTREGION_THRESHOLD_VERY_SUCCESSFUL,
):
    r"""Minimize a function with least squares structure using DFO-LS.

    For details see :ref:`list_of_nag_algorithms`.

    """
    if not IS_DFOLS_INSTALLED:
        raise NotImplementedError(
            "The dfols package is not installed and required for 'nag_dfols'. "
            "You can install it with 'pip install DFO-LS'. "
            "For additional installation instructions visit: ",
            r"https://numericalalgorithmsgroup.github.io/dfols/build/html/install.html",
        )
    if trustregion_method_to_replace_extra_points == "momentum":
        trustregion_use_momentum = True
    elif trustregion_method_to_replace_extra_points in ["geometry_improving", None]:
        trustregion_use_momentum = False
    else:
        raise ValueError(
            "trustregion_method_to_replace_extra_points must be "
            "'geometry_improving', 'momentum' or None."
        )

    algo_info = {
        "name": "nag_dfols",
        "primary_criterion_entry": "root_contributions",
        # does not really parallelize but we want to disable caching for noise averaging
        "parallelizes": True,
        "needs_scaling": False,
    }

    advanced_options, trustregion_reset_options = _create_nag_advanced_options(
        x=x,
        noise_multiplicative_level=noise_multiplicative_level,
        noise_additive_level=noise_additive_level,
        noise_n_evals_per_point=noise_n_evals_per_point,
        convergence_noise_corrected_criterion_tolerance=convergence_noise_corrected_criterion_tolerance,  # noqa: E501
        trustregion_initial_radius=trustregion_initial_radius,
        trustregion_reset_options=trustregion_reset_options,
        convergence_slow_progress=convergence_slow_progress,
        interpolation_rounding_error=interpolation_rounding_error,
        threshold_for_safety_step=threshold_for_safety_step,
        clip_criterion_if_overflowing=clip_criterion_if_overflowing,
        initial_directions=initial_directions,
        random_directions_orthogonal=random_directions_orthogonal,
        trustregion_precondition_interpolation=trustregion_precondition_interpolation,
        trustregion_threshold_successful=trustregion_threshold_successful,
        trustregion_threshold_very_successful=trustregion_threshold_very_successful,
        trustregion_shrinking_factor_not_successful=trustregion_shrinking_factor_not_successful,  # noqa: E501
        trustregion_expansion_factor_successful=trustregion_expansion_factor_successful,
        trustregion_expansion_factor_very_successful=trustregion_expansion_factor_very_successful,  # noqa: E501
        trustregion_shrinking_factor_lower_radius=trustregion_shrinking_factor_lower_radius,  # noqa: E501
        trustregion_shrinking_factor_upper_radius=trustregion_shrinking_factor_upper_radius,  # noqa: E501
    )

    fast_start = _build_options_dict(
        user_input=trustregion_fast_start_options,
        default_options=TRUSTREGION_FAST_START_OPTIONS,
    )
    if fast_start["floor_of_jacobian_singular_values"] != 1:
        warnings.warn(
            "Setting the `floor_of_jacobian_singular_values` is not supported by "
            "DF-OLS as of version 1.2.1."
        )
    if (
        fast_start["shrink_upper_radius_in_safety_steps"]
        and fast_start["full_geometry_improving_step"]
    ):
        raise ValueError(
            "full_geometry_improving_step of the trustregion_fast_start_options can "
            "only be True if shrink_upper_radius_in_safety_steps is False."
        )

    (
        faststart_jac,
        faststart_step,
    ) = _get_fast_start_method(fast_start["method"])

    if (
        trustregion_reset_options["n_extra_interpolation_points_per_soft_reset"]
        < trustregion_reset_options["n_extra_interpolation_points_per_soft_reset"]
    ):
        raise ValueError(
            "In the trustregion_reset_options "
            "'n_extra_interpolation_points_per_soft_reset' must "
            "be larger or the same as 'n_extra_interpolation_points_per_hard_reset'."
        )

    dfols_options = {
        "growing.full_rank.use_full_rank_interp": faststart_jac,
        "growing.perturb_trust_region_step": faststart_step,
        "restarts.hard.use_old_rk": trustregion_reset_options[
            "reuse_criterion_value_at_hard_reset"
        ],
        "restarts.auto_detect.min_chgJ_slope": trustregion_reset_options[
            "auto_detect_min_jacobian_increase"
        ],  # noqa: E501
        "restarts.max_npt": trustregion_reset_options["max_interpolation_points"],
        "restarts.increase_npt": trustregion_reset_options[
            "n_extra_interpolation_points_per_soft_reset"
        ]
        > 0,
        "restarts.increase_npt_amt": trustregion_reset_options[
            "n_extra_interpolation_points_per_soft_reset"
        ],
        "restarts.hard.increase_ndirs_initial_amt": trustregion_reset_options[
            "n_extra_interpolation_points_per_hard_reset"
        ]
        - trustregion_reset_options["n_extra_interpolation_points_per_soft_reset"],
        "model.rel_tol": convergence_scaled_criterion_tolerance,
        "regression.num_extra_steps": trustregion_n_extra_points_to_replace_successful,
        "regression.momentum_extra_steps": trustregion_use_momentum,
        "regression.increase_num_extra_steps_with_restart": trustregion_reset_options[
            "n_additional_extra_points_to_replace_per_reset"
        ],
        "growing.ndirs_initial": fast_start["min_inital_points"],
        "growing.delta_scale_new_dirns": fast_start[
            "scale_of_trustregion_step_perturbation"
        ],
        "growing.full_rank.scale_factor": fast_start[
            "scale_of_jacobian_components_perturbation"
        ],
        "growing.full_rank.svd_max_jac_cond": fast_start[
            "jacobian_max_condition_number"
        ],
        "growing.do_geom_steps": fast_start["geometry_improving_steps"],
        "growing.safety.do_safety_step": fast_start["safety_steps"],
        "growing.safety.reduce_delta": fast_start[
            "shrink_upper_radius_in_safety_steps"
        ],
        "growing.safety.full_geom_step": fast_start["full_geometry_improving_step"],
        "growing.reset_delta": fast_start["reset_trustregion_radius_after_fast_start"],
        "growing.reset_rho": fast_start[
            "reset_min_trustregion_radius_after_fast_start"
        ],
        "growing.gamma_dec": fast_start["shrinking_factor_not_successful"],
        "growing.num_new_dirns_each_iter": fast_start[
            "n_extra_search_directions_per_iteration"
        ],
    }

    advanced_options.update(dfols_options)

    criterion = partial(
        criterion_and_derivative, task="criterion", algorithm_info=algo_info
    )

    res = dfols.solve(
        criterion,
        x0=x,
        bounds=(lower_bounds, upper_bounds),
        maxfun=stopping_max_criterion_evaluations,
        rhobeg=trustregion_initial_radius,
        npt=trustregion_n_interpolation_points,
        rhoend=convergence_minimal_trustregion_radius_tolerance,
        nsamples=noise_n_evals_per_point,
        objfun_has_noise=noise_additive_level or noise_multiplicative_level,
        scaling_within_bounds=False,
        do_logging=False,
        print_progress=False,
        user_params=advanced_options,
    )

    return _process_nag_result(res, len(x))


def nag_pybobyqa(
    criterion_and_derivative,
    x,
    lower_bounds,
    upper_bounds,
    *,
    clip_criterion_if_overflowing=CLIP_CRITERION_IF_OVERFLOWING,
    convergence_criterion_value=None,
    convergence_minimal_trustregion_radius_tolerance=CONVERGENCE_MINIMAL_TRUSTREGION_RADIUS_TOLERANCE,  # noqa: E501
    convergence_noise_corrected_criterion_tolerance=CONVERGENCE_NOISE_CORRECTED_CRITERION_TOLERANCE,  # noqa: E501
    convergence_slow_progress=None,
    initial_directions=INITIAL_DIRECTIONS,
    interpolation_rounding_error=INTERPOLATION_ROUNDING_ERROR,
    noise_additive_level=None,
    noise_multiplicative_level=None,
    noise_n_evals_per_point=None,
    random_directions_orthogonal=RANDOM_DIRECTIONS_ORTHOGONAL,
    seek_global_optimum=False,
    stopping_max_criterion_evaluations=STOPPING_MAX_CRITERION_EVALUATIONS,
    threshold_for_safety_step=THRESHOLD_FOR_SAFETY_STEP,
    trustregion_expansion_factor_successful=TRUSTREGION_EXPANSION_FACTOR_SUCCESSFUL,
    trustregion_expansion_factor_very_successful=TRUSTREGION_EXPANSION_FACTOR_VERY_SUCCESSFUL,  # noqa: E501
    trustregion_initial_radius=None,
    trustregion_minimum_change_hession_for_underdetermined_interpolation=True,
    trustregion_n_interpolation_points=None,
    trustregion_precondition_interpolation=TRUSTREGION_PRECONDITION_INTERPOLATION,
    trustregion_reset_options=None,
    trustregion_shrinking_factor_not_successful=TRUSTREGION_SHRINKING_FACTOR_NOT_SUCCESSFUL,  # noqa: E501
    trustregion_shrinking_factor_lower_radius=TRUSTREGION_SHRINKING_FACTOR_LOWER_RADIUS,
    trustregion_shrinking_factor_upper_radius=TRUSTREGION_SHRINKING_FACTOR_UPPER_RADIUS,
    trustregion_threshold_successful=TRUSTREGION_THRESHOLD_SUCCESSFUL,
    trustregion_threshold_very_successful=TRUSTREGION_THRESHOLD_VERY_SUCCESSFUL,
):
    r"""Minimize a function using the BOBYQA algorithm.

    For details see :ref:`list_of_nag_algorithms`.

    """
    if not IS_PYBOBYQA_INSTALLED:
        raise NotImplementedError(
            "The pybobyqa package is not installed and required for 'nag_pybobyqa'. "
            "You can install it with 'pip install Py-BOBYQA'. "
            "For additional installation instructions visit: ",
            r"https://numericalalgorithmsgroup.github.io/pybobyqa/build/html/"
            "install.html",
        )

    if convergence_criterion_value is None:
        convergence_criterion_value = -np.inf

    algo_info = {
        "name": "nag_pybobyqa",
        "primary_criterion_entry": "value",
        # does not really parallelize but we want to disable caching for noise averaging
        "parallelizes": True,
        "needs_scaling": False,
    }
    criterion = partial(
        criterion_and_derivative, task="criterion", algorithm_info=algo_info
    )
    advanced_options, trustregion_reset_options = _create_nag_advanced_options(
        x=x,
        noise_multiplicative_level=noise_multiplicative_level,
        noise_additive_level=noise_additive_level,
        trustregion_initial_radius=trustregion_initial_radius,
        noise_n_evals_per_point=noise_n_evals_per_point,
        convergence_noise_corrected_criterion_tolerance=convergence_noise_corrected_criterion_tolerance,  # noqa: E501
        trustregion_reset_options=trustregion_reset_options,
        convergence_slow_progress=convergence_slow_progress,
        interpolation_rounding_error=interpolation_rounding_error,
        threshold_for_safety_step=threshold_for_safety_step,
        clip_criterion_if_overflowing=clip_criterion_if_overflowing,
        initial_directions=initial_directions,
        random_directions_orthogonal=random_directions_orthogonal,
        trustregion_precondition_interpolation=trustregion_precondition_interpolation,
        trustregion_threshold_successful=trustregion_threshold_successful,
        trustregion_threshold_very_successful=trustregion_threshold_very_successful,
        trustregion_shrinking_factor_not_successful=trustregion_shrinking_factor_not_successful,  # noqa: E501
        trustregion_expansion_factor_successful=trustregion_expansion_factor_successful,
        trustregion_expansion_factor_very_successful=trustregion_expansion_factor_very_successful,  # noqa: E501
        trustregion_shrinking_factor_lower_radius=trustregion_shrinking_factor_lower_radius,  # noqa: E501
        trustregion_shrinking_factor_upper_radius=trustregion_shrinking_factor_upper_radius,  # noqa: E501
    )

    pybobyqa_options = {
        "model.abs_tol": convergence_criterion_value,
        "interpolation.minimum_change_hessian": trustregion_minimum_change_hession_for_underdetermined_interpolation,  # noqa: E501
        "restarts.max_unsuccessful_restarts_total": trustregion_reset_options[
            "max_unsuccessful_resets"
        ],
        "restarts.rhobeg_scale_after_unsuccessful_restart": trustregion_reset_options[
            "trust_region_scaling_at_unsuccessful_reset"
        ],
        "restarts.hard.use_old_fk": trustregion_reset_options[
            "reuse_criterion_value_at_hard_reset"
        ],
        "restarts.auto_detect.min_chg_model_slope": trustregion_reset_options[
            "auto_detect_min_jacobian_increase"
        ],
    }

    advanced_options.update(pybobyqa_options)

    res = pybobyqa.solve(
        criterion,
        x0=x,
        bounds=(lower_bounds, upper_bounds),
        maxfun=stopping_max_criterion_evaluations,
        rhobeg=trustregion_initial_radius,
        user_params=advanced_options,
        scaling_within_bounds=False,
        do_logging=False,
        print_progress=False,
        objfun_has_noise=noise_additive_level or noise_multiplicative_level,
        nsamples=noise_n_evals_per_point,
        npt=trustregion_n_interpolation_points,
        rhoend=convergence_minimal_trustregion_radius_tolerance,
        seek_global_minimum=seek_global_optimum,
    )

    return _process_nag_result(res, len(x))


def _process_nag_result(nag_result_obj, len_x):
    """Convert the NAG result object to our result dictionary.

    Args:
        nag_result_obj: NAG result object
        len_x (int): length of the supplied parameters, i.e. the dimensionality of the
            problem.


    Returns:
        results (dict): See :ref:`internal_optimizer_output` for details.
    """
    processed = {
        "solution_criterion": nag_result_obj.f,
        "n_criterion_evaluations": nag_result_obj.nx,
        "message": nag_result_obj.msg,
        "success": nag_result_obj.flag == nag_result_obj.EXIT_SUCCESS,
        "reached_convergence_criterion": None,
    }
    if nag_result_obj.x is not None:
        processed["solution_x"] = nag_result_obj.x
    else:
        processed["solution_x"] = np.array([np.nan] * len_x)
    try:
        processed["solution_derivative"] = nag_result_obj.gradient
    except AttributeError:
        pass
    try:
        processed["solution_hessian"] = nag_result_obj.hessian
    except AttributeError:
        pass
    if processed["message"].startswith("Error (bad input)"):
        raise ValueError(processed["message"])
    return processed


def _create_nag_advanced_options(
    x,
    noise_multiplicative_level,
    noise_additive_level,
    trustregion_initial_radius,
    noise_n_evals_per_point,
    convergence_noise_corrected_criterion_tolerance,
    trustregion_reset_options,
    convergence_slow_progress,
    interpolation_rounding_error,
    threshold_for_safety_step,
    clip_criterion_if_overflowing,
    initial_directions,
    random_directions_orthogonal,
    trustregion_precondition_interpolation,
    trustregion_threshold_successful,
    trustregion_threshold_very_successful,
    trustregion_shrinking_factor_not_successful,
    trustregion_expansion_factor_successful,
    trustregion_expansion_factor_very_successful,
    trustregion_shrinking_factor_lower_radius,
    trustregion_shrinking_factor_upper_radius,
):
    if noise_multiplicative_level is not None and noise_additive_level is not None:
        raise ValueError("You cannot specify both multiplicative and additive noise.")
    if trustregion_initial_radius is None:
        trustregion_initial_radius = calculate_trustregion_initial_radius(x)
    # -np.inf as a default leads to errors when building the documentation with sphinx.
    noise_n_evals_per_point = _change_evals_per_point_interface(noise_n_evals_per_point)
    trustregion_reset_options = _build_options_dict(
        user_input=trustregion_reset_options,
        default_options=RESET_OPTIONS,
    )
    if trustregion_reset_options["reset_type"] not in ["soft", "hard"]:
        raise ValueError(
            "reset_type in the trustregion_reset_options must be soft or hard."
        )
    if initial_directions not in ["coordinate", "random"]:
        raise ValueError("inital_directions must be either 'coordinate' or 'random'.")
    convergence_slow_progress = _build_options_dict(
        user_input=convergence_slow_progress,
        default_options=CONVERGENCE_SLOW_PROGRESS,
    )

    is_noisy = bool(noise_additive_level or noise_multiplicative_level)

    advanced_options = {
        "general.rounding_error_constant": interpolation_rounding_error,
        "general.safety_step_thresh": threshold_for_safety_step,
        "general.check_objfun_for_overflow": clip_criterion_if_overflowing,
        "tr_radius.eta1": trustregion_threshold_successful,
        "tr_radius.eta2": trustregion_threshold_very_successful,
        "tr_radius.gamma_dec": trustregion_shrinking_factor_not_successful,
        "tr_radius.gamma_inc": trustregion_expansion_factor_successful,
        "tr_radius.gamma_inc_overline": trustregion_expansion_factor_very_successful,
        "tr_radius.alpha1": trustregion_shrinking_factor_lower_radius,
        "tr_radius.alpha2": trustregion_shrinking_factor_upper_radius,
        "general.rounding_error_constant": interpolation_rounding_error,
        "general.safety_step_thresh": threshold_for_safety_step,
        "general.check_objfun_for_overflow": clip_criterion_if_overflowing,
        "init.random_initial_directions": initial_directions == "random",
        "init.random_directions_make_orthogonal": random_directions_orthogonal,
        "slow.thresh_for_slow": convergence_slow_progress[
            "threshold_to_characterize_as_slow"
        ],
        "slow.max_slow_iters": convergence_slow_progress[
            "max_insufficient_improvements"
        ],
        "slow.history_for_slow": convergence_slow_progress["comparison_period"],
        "noise.multiplicative_noise_level": noise_multiplicative_level,
        "noise.additive_noise_level": noise_additive_level,
        "noise.quit_on_noise_level": (
            convergence_noise_corrected_criterion_tolerance > 0
        )
        and is_noisy,
        "noise.scale_factor_for_quit": convergence_noise_corrected_criterion_tolerance,
        "interpolation.precondition": trustregion_precondition_interpolation,
        "restarts.use_restarts": trustregion_reset_options["use_resets"],
        "restarts.max_unsuccessful_restarts": trustregion_reset_options[
            "max_consecutive_unsuccessful_resets"
        ],
        "restarts.rhoend_scale": trustregion_reset_options[
            "minimal_trustregion_radius_tolerance_scaling_at_reset"
        ],
        "restarts.use_soft_restarts": trustregion_reset_options["reset_type"] == "soft",
        "restarts.soft.move_xk": trustregion_reset_options["move_center_at_soft_reset"],
        "restarts.soft.max_fake_successful_steps": trustregion_reset_options[
            "max_iterations_without_new_best_after_soft_reset"
        ],  # noqa: E501
        "restarts.auto_detect": trustregion_reset_options["auto_detect"],
        "restarts.auto_detect.history": trustregion_reset_options[
            "auto_detect_history"
        ],  # noqa: E501
        "restarts.auto_detect.min_correl": trustregion_reset_options[
            "auto_detect_min_correlations"
        ],
        "restarts.soft.num_geom_steps": trustregion_reset_options[
            "points_to_replace_at_soft_reset"
        ],
    }

    return advanced_options, trustregion_reset_options


def _change_evals_per_point_interface(func):
    """Change the interface of the user supplied function to the one expected
    by NAG.

    Args:
        func (callable or None): function mapping from our names to
            noise_n_evals_per_point.

    Returns:
        adjusted_noise_n_evals_per_point (callable): function mapping from the
            argument names expected by pybobyqa and df-ols to noise_n_evals_per_point.
    """
    if func is not None:

        def adjusted_noise_n_evals_per_point(delta, rho, iter, nrestarts):  # noqa: A002
            return func(
                upper_trustregion_radius=delta,
                lower_trustregion_radius=rho,
                n_iterations=iter,
                n_resets=nrestarts,
            )

        return adjusted_noise_n_evals_per_point


def _build_options_dict(user_input, default_options):
    """Create the full dictionary of trust region fast start options from user input.

    Args:
        user_input (dict or None): dictionary to update the default options with.
            May only contain keys present in the default options.
        default_options (dict): the default values.

    Returns:
        full_options (dict)

    """
    full_options = default_options.copy()
    user_input = {} if user_input is None else user_input
    invalid = [x for x in user_input if x not in full_options]
    if len(invalid) > 0:
        raise ValueError(
            f"You specified illegal options {', '.join(invalid)}. Allowed are: "
            ", ".join(full_options.keys())
        )
    full_options.update(user_input)
    return full_options


def _get_fast_start_method(user_value):
    """Get fast start method arguments from user value."""
    allowed_values = ["auto", "jacobian", "trustregion"]
    if user_value not in allowed_values:
        raise ValueError(
            "`perturb_jacobian_or_trustregion_step` must be one of "
            f"{allowed_values}. You provided {user_value}."
        )
    if user_value == "auto":
        faststart_jac = None
        faststart_step = None
    else:
        faststart_jac = user_value == "jacobian"
        faststart_step = not faststart_jac

    return faststart_jac, faststart_step



"""Functions to read data from the database used for logging.

The functions in the module are meant for end users of estimagic.
They do not require any knowledge of databases.

When using them internally (e.g. in the dashboard), make sure to supply a database to
path_or_database. Otherwise, the functions may be very slow.

"""
from pathlib import Path

import pandas as pd
from estimagic.logging.database_utilities import load_database
from estimagic.logging.database_utilities import read_last_rows
from estimagic.logging.database_utilities import read_new_rows
from estimagic.logging.database_utilities import read_specific_row
from sqlalchemy import MetaData


def read_optimization_iteration(path_or_database, iteration, include_internals=False):
    """Get information about an optimization iteration.

    Args:
        path_or_database (pathlib.Path, str or sqlalchemy.MetaData)
        iteration (int): The index of the iteration that should be retrieved.
            The row_id behaves as Python list indices, i.e. ``0`` identifies the
            first iteration, ``-1`` the last one, etc.
        include_internals (bool): Whether internally used quantities like the
            internal parameter vector and the corresponding derivative etc. are included
            in the result. Default False. This should only be used by advanced users.

    Returns:
        dict: The logged information corresponding to the iteration. The keys correspond
            to database columns.

    Raises:
        KeyError: if the iteration is out of bounds.

    """
    database = load_database(**_process_path_or_database(path_or_database))
    start_params = read_start_params(database)
    if iteration >= 0:
        rowid = iteration + 1
    else:
        last_iteration = read_last_rows(
            database=database,
            table_name="optimization_iterations",
            n_rows=1,
            return_type="list_of_dicts",
        )
        highest_rowid = last_iteration[0]["rowid"]

        rowid = highest_rowid + iteration + 1

    data = read_specific_row(
        database=database,
        table_name="optimization_iterations",
        rowid=rowid,
        return_type="list_of_dicts",
    )

    if len(data) == 0:
        raise IndexError(f"Invalid iteration requested: {iteration}")
    else:
        data = data[0]

    params = start_params.copy()
    params["value"] = data.pop("params")
    data["params"] = params

    to_remove = ["distance_origin", "distance_ones"]
    if not include_internals:
        to_remove += ["internal_params", "internal_derivative"]
    for key in to_remove:
        if key in data:
            del data[key]

    return data


def read_start_params(path_or_database):
    """Load the start parameters DataFrame.

    Args:
        path_or_database (pathlib.Path, str or sqlalchemy.MetaData)

    Returns:
        params (pd.DataFrame): see :ref:`params`.

    """
    database = load_database(**_process_path_or_database(path_or_database))
    optimization_problem = read_last_rows(
        database=database,
        table_name="optimization_problem",
        n_rows=1,
        return_type="dict_of_lists",
    )
    start_params = optimization_problem["params"][0]
    return start_params


def read_optimization_histories(path_or_database):
    """Read a histories out values, parameters and other information."""
    database = load_database(**_process_path_or_database(path_or_database))

    start_params = read_start_params(path_or_database)

    raw_res, _ = read_new_rows(
        database=database,
        table_name="optimization_iterations",
        last_retrieved=0,
        return_type="dict_of_lists",
    )

    params_history = pd.DataFrame(raw_res["params"], columns=start_params.index)
    value_history = pd.Series(raw_res["value"])

    metadata = pd.DataFrame()
    metadata["timestamps"] = raw_res["timestamp"]
    metadata["valid"] = raw_res["valid"]
    metadata["has_value"] = value_history.notnull()
    metadata["has_derivative"] = [d is not None for d in raw_res["internal_derivative"]]

    histories = {
        "values": value_history.dropna(),
        "params": params_history,
        "metadata": metadata,
    }

    if "contributions" in raw_res:
        first_contrib = raw_res["contributions"][0]
        if isinstance(first_contrib, pd.Series):
            columns = first_contrib.index
        else:
            columns = None
        contributions_history = pd.DataFrame(
            raw_res["contributions"], columns=columns
        ).dropna()
        histories["contributions"] = contributions_history

    return histories


def _process_path_or_database(path_or_database):
    """Make inputs for load_database out of path_or_database.

    Args:
        path_or_database (pathlib.Path, str or sqlalchemy.MetaData)

    Returns:
        dict: The keys are "path", "metadata" and "fast_logging"

    Examples:

    >>> from sqlalchemy import MetaData
    >>> database = MetaData()
    >>> _process_path_or_database(database)
    {'path': None, 'metadata': MetaData(), 'fast_logging': False}

    """
    res = {"path": None, "metadata": None, "fast_logging": False}
    if isinstance(path_or_database, MetaData):
        res["metadata"] = path_or_database
    elif isinstance(path_or_database, (Path, str)):
        res["path"] = Path(path_or_database).resolve()
        if not res["path"].exists():
            raise FileNotFoundError(f"No such database file: {res['path']}")
    else:
        raise ValueError(
            "path_or_database must be a path or sqlalchemy.MetaData object"
        )
    return res


def read_steps_table(path_or_database):
    """Load the start parameters DataFrame.

    Args:
        path_or_database (pathlib.Path, str or sqlalchemy.MetaData)

    Returns:
        params (pd.DataFrame): see :ref:`params`.

    """
    database = load_database(**_process_path_or_database(path_or_database))
    steps_table, _ = read_new_rows(
        database=database,
        table_name="steps",
        last_retrieved=0,
        return_type="list_of_dicts",
    )
    steps_df = pd.DataFrame(steps_table)

    return steps_df



"""Implementation of the Steihaug-Toint Conjugate Gradient algorithm."""
from math import sqrt

import numpy as np


def minimize_trust_cg(model_gradient, model_hessian, trustregion_radius):
    """Minimize the quadratic trust-region subproblem using Conjugate Gradient method.

    Args:
        model_gradient (np.ndarray): Gradient of the quadratic model. Shape (n,).
        model_hessian (np.ndarray): Hessian of the quadratic model. Shape (n, n).
        trustregion_radius (float): Radius of the trust-region.

    Returns:
        (np.ndarray): Solution vector of shape (n,).
    """
    n = model_gradient.shape[0]
    maxiter = 2 * n

    gtol_abs = 1e-8
    gtol_rel = 1e-6

    residual = np.copy(model_gradient)
    x_candidate = np.zeros(n)

    # Use steepest descent direction at initial point
    direction = -residual

    gradient_norm = np.linalg.norm(residual)
    stop_tol = max(gtol_abs, gtol_rel * gradient_norm)

    for _niter in range(maxiter):

        if gradient_norm <= stop_tol:
            break

        square_terms = np.dot(np.dot(direction, model_hessian), direction)

        distance_to_boundary = _get_distance_to_trustregion_boundary(
            x_candidate, direction, trustregion_radius
        )

        # Length of the Conjugate Gradient step
        alpha = np.dot(residual, residual) / square_terms

        if square_terms <= 0 or alpha > distance_to_boundary:
            x_candidate = x_candidate + distance_to_boundary * direction
            break

        x_candidate, residual, direction = _update_vectors_for_next_iteration(
            x_candidate, residual, direction, model_hessian, alpha
        )
        gradient_norm = np.linalg.norm(residual)

    return x_candidate


def _update_vectors_for_next_iteration(
    x_candidate, residual, direction, hessian, alpha
):
    """Update candidate, residual, and direction vectors for the next iteration.

    Args:
        x_candidate (np.ndarray): Candidate vector of shape (n,).
        residual (np.ndarray): Array of residuals of shape (n,). The residual vector
            is defined as `r = Ax - b`, where `A` denotes the hessian matrix and `b` the
            gradient vector of the quadratic trust-region subproblem.
            `r` is equivalent to the first derivative of the quadratic subproblem.
        direction (np.ndarray): Direction vector of shape (n,).

    Returns:
        (tuple) Tuple containing:
            - x_candidate (np.ndarray): Updated candidate vector of shape (n,).
            - residual (np.ndarray): Updated array of residuals of shape (n,).
            - direction (np.darray): Updated direction vector of shape (n,).
    """
    residual_old = np.copy(residual)

    x_candidate = x_candidate + alpha * direction
    residual = residual + alpha * np.dot(hessian, direction)

    beta = np.dot(residual, residual) / np.dot(residual_old, residual_old)
    direction = -residual + beta * direction

    return x_candidate, residual, direction


def _get_distance_to_trustregion_boundary(candidate, direction, radius):
    """Compute the distance of the candidate vector to trustregion boundary.

    The positive distance sigma is defined in Eculidean norm, as follows:

        `|| s + sigma * p || = radius`

    where `s` denotes the candidate vector, and `p` the direction vector.

    Args:
        candidate(np.ndarray): Candidate vector of shape (n,).
        direction (np.ndarray): Direction vector of shape (n,).
        radius (floar): Radius of the trust-region

    Returns:
        (float) Distance of the candidate vector to the trustregion
            boundary.
    """
    ss = np.dot(candidate, candidate)
    sp = np.dot(candidate, direction)
    pp = np.dot(direction, direction)

    sigma = -sp + sqrt(sp * sp + pp * (radius**2 - ss))
    sigma /= pp

    return sigma



import functools
from collections.abc import Callable

import numpy as np
import pandas as pd
import scipy
from estimagic.decorators import numpy_interface
from estimagic.differentiation.derivatives import first_derivative
from estimagic.parameters.parameter_conversion import get_internal_bounds
from estimagic.parameters.parameter_conversion import get_reparametrize_functions
from estimagic.parameters.process_constraints import process_constraints


def transform_covariance(
    params,
    internal_cov,
    constraints,
    n_samples,
    bounds_handling,
):
    """Transform the internal covariance matrix to an external one, given constraints.

    Args:
        params (pd.DataFrame): DataFrame where the "value" column contains estimated
            parameters of a likelihood model. See :ref:`params` for details.
        internal_cov (np.ndarray or pandas.DataFrame) with a covariance matrix of the
            internal parameter vector. For background information about internal and
            external params see :ref:`implementation_of_constraints`.
        constraints (list): List with constraint dictionaries.
            See .. _link: ../../docs/source/how_to_guides/how_to_use_constraints.ipynb
        n_samples (int): Number of samples used to transform the covariance matrix of
            the internal parameter vector into the covariance matrix of the external
            parameters.
        bounds_handling (str): One of "clip", "raise", "ignore". Determines how bounds
            are handled. If "clip", confidence intervals are clipped at the bounds.
            Standard errors are only adjusted if a sampling step is necessary due to
            additional constraints. If "raise" and any lower or upper bound is binding,
            we raise an error. If "ignore", boundary problems are simply ignored.

    Returns:
        pd.DataFrame: Quadratic DataFrame containing the covariance matrix of the free
            parameters. If parameters were fixed (explicitly or by other constraints),
            the index is a subset of params.index. The columns are the same as the
            index.

    """
    processed_constraints, processed_params = process_constraints(constraints, params)
    free_index = processed_params.query("_internal_free").index

    if isinstance(internal_cov, pd.DataFrame):
        internal_cov = internal_cov.to_numpy()

    if processed_constraints:
        _to_internal, _from_internal = get_reparametrize_functions(
            params=params, constraints=constraints
        )

        free = processed_params.loc[free_index]
        is_free = processed_params["_internal_free"].to_numpy()
        lower_bounds = free["_internal_lower"]
        upper_bounds = free["_internal_upper"]

        internal_mean = _to_internal(params)

        sample = np.random.multivariate_normal(
            mean=internal_mean,
            cov=internal_cov,
            size=n_samples,
        )
        transformed_free = []
        for params_vec in sample:
            if bounds_handling == "clip":
                params_vec = np.clip(params_vec, a_min=lower_bounds, a_max=upper_bounds)
            elif bounds_handling == "raise":
                if (params_vec < lower_bounds).any() or (
                    params_vec > upper_bounds
                ).any():
                    raise ValueError()

            transformed = _from_internal(internal=params_vec)
            transformed_free.append(transformed[is_free])

        free_cov = np.cov(
            np.array(transformed_free),
            rowvar=False,
        )

    else:
        free_cov = internal_cov

    res = pd.DataFrame(data=free_cov, columns=free_index, index=free_index)
    return res


def calculate_inference_quantities(params, free_cov, ci_level):
    """Add standard errors, pvalues and confidence intervals to params.

    Args
        params (pd.DataFrame): See :ref:`params`.
        free_cov (pd.DataFrame): Quadratic DataFrame containing the covariance matrix
            of the free parameters. If parameters were fixed (explicitly or by other
            constraints) the index is a subset of params.index. The columns are the same
            as the index.
        ci_level (float): Confidence level for the calculation of confidence intervals.

    Returns:
        pd.DataFrame: DataFrame with same index as params, containing the columns
            "value", "standard_error", "pvalue", "ci_lower" and "ci_upper".
            Parameters that do not have a standard error (e.g. because they were fixed
            during estimation) contain NaNs in all but the "value" column. The value
            column is only reproduced for convenience.

    """
    free = pd.DataFrame(index=free_cov.index)
    free["value"] = params.loc[free.index, "value"]
    free["standard_error"] = np.sqrt(np.diag(free_cov))
    tvalues = free["value"] / free["standard_error"]
    free["p_value"] = 1.96 * scipy.stats.norm.sf(np.abs(tvalues))

    alpha = 1 - ci_level
    scale = scipy.stats.norm.ppf(1 - alpha / 2)
    free["ci_lower"] = free["value"] - scale * free["standard_error"]
    free["ci_upper"] = free["value"] + scale * free["standard_error"]

    free["stars"] = pd.cut(
        free["p_value"], bins=[-1, 0.01, 0.05, 0.1, 2], labels=["***", "**", "*", ""]
    )

    res = free.reindex(params.index)
    res["value"] = params["value"]
    return res


def get_internal_first_derivative(
    func, params, constraints=None, func_kwargs=None, numdiff_options=None
):
    """Get the first_derivative of func with respect to internal parameters.

    If there are no constraints, we simply call the first_derivative function.

    Args:
        func (callable): Function to take the derivative of.
        params (pandas.DataFrame): Data frame with external parameters. See
            :ref:`params`.
        constraints (list): Constraints that define how to convert between internal
            and external parameters.
        func_kwargs (dict): Additional keyword arguments for func.
        numdiff_options (dict): Additional options for first_derivative.

    Returns:
        dict: See ``first_derivative`` for details. The only difference is that the
            the "derivative" entry is always a numpy array instead of a DataFrame

    """
    numdiff_options = {} if numdiff_options is None else numdiff_options
    func_kwargs = {} if func_kwargs is None else func_kwargs
    _func = functools.partial(func, **func_kwargs)

    if constraints is None:
        out = first_derivative(
            func=_func,
            params=params,
            **numdiff_options,
        )
        out["has_transforming_constraints"] = False
    else:

        lower_bounds, upper_bounds = get_internal_bounds(params, constraints)

        _internal_func = numpy_interface(
            func=_func, params=params, constraints=constraints
        )

        _to_internal, _ = get_reparametrize_functions(params, constraints)

        _x = _to_internal(params)

        out = first_derivative(
            _internal_func,
            _x,
            lower_bounds=lower_bounds,
            upper_bounds=upper_bounds,
            **numdiff_options,
        )

        if isinstance(out["derivative"], (pd.DataFrame, pd.Series)):
            out["derivative"] = out["derivative"].to_numpy()

    return out


def process_pandas_arguments(**kwargs):
    """Convert pandas objects to arrays and extract names of moments and parameters.

    This works for any number of keyword arguments. The result is a tuple containing
    numpy arrays in same order as the keyword arguments and a dictionary with
    the separated index objects as last entry. This dictionary contains the entries
    "moments" and "params" for the identified moment names and parameter names.

    The keyword arguments "jac", "hess", "weights" and "moments_cov" are used to extract
    the names. Other keyword arguments are simply converted to numpy arrays.

    """
    param_name_candidates = {}
    moment_name_candidates = {}

    if "jac" in kwargs:
        jac = kwargs["jac"]
        if isinstance(jac, pd.DataFrame):
            param_name_candidates["jac"] = jac.columns
            moment_name_candidates["jac"] = jac.index

    if "hess" in kwargs:
        hess = kwargs["hess"]
        if isinstance(hess, pd.DataFrame):
            param_name_candidates["hess"] = hess.index

    if "weights" in kwargs:
        weights = kwargs["weights"]
        if isinstance(weights, pd.DataFrame):
            moment_name_candidates["weights"] = weights.index

    if "moments_cov" in kwargs:
        moments_cov = kwargs["moments_cov"]
        if isinstance(moments_cov, pd.DataFrame):
            moment_name_candidates["moments_cov"] = moments_cov.index

    names = {}
    if param_name_candidates:
        _check_names_coincide(param_name_candidates)
        names["params"] = list(param_name_candidates.values())[0]
    if moment_name_candidates:
        _check_names_coincide(moment_name_candidates)
        names["moments"] = list(moment_name_candidates.values())[0]

    # order of outputs is same as order of inputs; names are last.
    out_list = [_to_numpy(val, name=key) for key, val in kwargs.items()] + [names]
    return tuple(out_list)


def _to_numpy(df_or_array, name):
    if isinstance(df_or_array, pd.DataFrame):
        arr = df_or_array.to_numpy()
    elif isinstance(df_or_array, np.ndarray):
        arr = df_or_array
    else:
        raise ValueError(
            f"{name} must be a DataFrame or numpy array, not {type(df_or_array)}."
        )
    return arr


def _check_names_coincide(name_dict):
    if len(name_dict) >= 2:
        first_key = list(name_dict)[0]
        first_names = name_dict[first_key]

        for key, names in name_dict.items():
            if not first_names.equals(names):
                msg = f"Ambiguous parameter or moment names from {first_key} and {key}."
                raise ValueError(msg)


def get_derivative_case(derivative):
    """Determine which kind of derivative should be used."""
    if isinstance(derivative, (pd.DataFrame, np.ndarray)):
        case = "pre-calculated"
    elif isinstance(derivative, Callable):
        case = "closed-form"
    elif derivative is False:
        case = "skip"
    else:
        case = "numerical"
    return case


def check_is_optimized_and_derivative_case(is_minimized, derivative_case):
    if (not is_minimized) and derivative_case == "pre-calculated":
        raise ValueError(
            "Providing a pre-calculated derivative is only possible if the "
            "optimization was done outside of the estimate_function, i.e. if "
            "optimize_options=False."
        )



"""Auxiliary functions for the quadratic BNTR algorithm."""
from collections import namedtuple
from copy import copy
from functools import reduce

import numpy as np
from estimagic.optimization._trustregion_conjugate_gradient_quadratic import (
    minimize_trust_cg,
)

EPSILON = np.finfo(float).eps ** (2 / 3)


def take_preliminary_steepest_descent_step_and_check_for_solution(
    x_candidate,
    model,
    lower_bounds,
    upper_bounds,
    maxiter_steepest_descent,
    step_size_newton,
    gtol_abs,
    gtol_rel,
    gtol_scaled,
):
    """Take a preliminary steepest descent step and check if we find a solution."""
    f_candidate = evaluate_model_criterion(
        x_candidate, model.linear_terms, model.square_terms
    )

    newton_step = compute_newton_step(model)
    active_bounds_info = get_information_on_active_bounds(
        x_candidate, lower_bounds, upper_bounds, model, newton_step, step_size_newton
    )

    gradient_candidate = evaluate_model_gradient(x_candidate, model)
    gradient_projected = np.where(
        np.arange(x_candidate.shape[0]) == active_bounds_info.all, 0, gradient_candidate
    )

    converged = check_for_convergence_steepest_descent(
        x_candidate,
        f_candidate,
        gradient_candidate,
        model,
        lower_bounds,
        upper_bounds,
        gtol_abs,
        gtol_rel,
        gtol_scaled,
    )

    if converged is True:
        x_accepted = x_candidate
        f_accepted = f_candidate

        hessian_inactive = np.copy(model.square_terms)
        trustregion_radius = 100
    else:
        hessian_inactive = find_hessian_submatrix_where_bounds_inactive(
            model, active_bounds_info
        )

        (
            x_candidate_steepest_descent,
            f_min_steepest_descent,
            trustregion_radius,
            max_trustregion_radius,
        ) = perform_steepest_descent_and_update_trustregion_radius(
            x_candidate,
            f_candidate,
            gradient_projected,
            hessian_inactive,
            model,
            lower_bounds,
            upper_bounds,
            active_bounds_info,
            maxiter_steepest_descent,
        )

        if f_min_steepest_descent < f_candidate:
            x_accepted = x_candidate_steepest_descent
            f_accepted = f_min_steepest_descent

            gradient_candidate_steepest_descent = evaluate_model_gradient(
                x_candidate_steepest_descent, model
            )

            converged = check_for_convergence_steepest_descent(
                x_candidate_steepest_descent,
                f_min_steepest_descent,
                gradient_candidate_steepest_descent,
                model,
                lower_bounds,
                upper_bounds,
                gtol_abs,
                gtol_rel,
                gtol_scaled,
            )
        else:
            x_accepted = x_candidate
            f_accepted = f_candidate

        trustregion_radius = max(trustregion_radius, max_trustregion_radius)
        trustregion_radius = min(max(trustregion_radius, 1e-10), 1e10)

    return (
        x_accepted,
        f_accepted,
        gradient_candidate,
        hessian_inactive,
        trustregion_radius,
        active_bounds_info,
        converged,
    )


def compute_conjugate_gradient_step(
    x,
    gradient_candidate,
    hessian_inactive,
    lower_bounds,
    upper_bounds,
    active_bounds_info,
    trustregion_radius,
):
    """Compute the bounded Conjugate Gradient trust-region step."""
    conjugate_gradient_step = np.zeros(x.shape[0])
    gradient_inactive_cg = np.copy(gradient_candidate[active_bounds_info.inactive])

    x_inactive_cg = minimize_trust_cg(
        gradient_inactive_cg,
        hessian_inactive,
        trustregion_radius,
    )
    gradient_norm_cg = np.linalg.norm(x_inactive_cg)

    conjugate_gradient_step[active_bounds_info.inactive] = -np.copy(x_inactive_cg)
    conjugate_gradient_step = _apply_bounds_to_conjugate_gradient_step(
        x,
        lower_bounds,
        upper_bounds,
        active_bounds_info,
        conjugate_gradient_step,
    )

    return (
        x_inactive_cg,
        gradient_inactive_cg,
        gradient_norm_cg,
        conjugate_gradient_step,
    )


def compute_predicted_reduction_from_conjugate_gradient_step(
    x_inactive_cg,
    gradient_inactive_cg,
    conjugate_gradient_step,
    gradient_candidate,
    hessian_inactive,
    active_bounds_info,
):
    """Compute predicted reduction induced by the Conjugate Gradient step."""
    if active_bounds_info.all.size > 0:
        # Projection changed the step, so we have to recompute the step
        # and the predicted reduction. Leave the rust radius unchanged.
        x_inactive_recomp = conjugate_gradient_step[active_bounds_info.inactive]
        gradient_inactive_recomp = gradient_candidate[active_bounds_info.inactive]

        predicted_reduction = evaluate_model_criterion(
            x_inactive_recomp, gradient_inactive_recomp, hessian_inactive
        )
    else:
        # Step did not change, so we can just recover the
        # pre-computed prediction
        predicted_reduction = evaluate_model_criterion(
            x_inactive_cg,
            gradient_inactive_cg,
            hessian_inactive,
        )

        predicted_reduction = -predicted_reduction

    return predicted_reduction


def perform_steepest_descent_and_update_trustregion_radius(
    x_candidate,
    f_candidate,
    gradient_projected,
    hessian_inactive,
    model,
    lower_bounds,
    upper_bounds,
    active_bounds_info,
    maxiter_steepest_descent,
):
    """Perform steepest descent step and update trust-region radius."""
    x_trial = np.copy(x_candidate)
    f_min = np.copy(f_candidate)

    step_size_accepted = 0
    trustregion_radius = 100
    max_trustregion_radius = 0

    gradient_norm = np.linalg.norm(gradient_projected)

    for _ in range(maxiter_steepest_descent):
        x_old = np.copy(x_trial)

        step_size_trial = trustregion_radius / gradient_norm
        x_trial = x_trial - step_size_trial * gradient_projected
        x_trial = apply_bounds_to_x_candidate(x_trial, lower_bounds, upper_bounds)
        f_trial = evaluate_model_criterion(
            x_trial, model.linear_terms, model.square_terms
        )

        x_accepted = x_trial - x_old

        if f_trial < f_min:
            f_min = copy(f_trial)
            step_size_accepted = trustregion_radius / gradient_norm

        x_inactive = np.copy(x_accepted[active_bounds_info.inactive])

        square_terms = np.dot(np.dot(x_inactive, hessian_inactive), x_inactive)

        predicted_reduction = trustregion_radius * (
            gradient_norm
            - 0.5 * trustregion_radius * square_terms / (gradient_norm**2)
        )
        actual_reduction = f_candidate - f_trial

        (
            tau,
            max_trustregion_radius,
        ) = _update_tau_and_max_trustregion_radius_steepest_descent(
            trustregion_radius,
            predicted_reduction,
            actual_reduction,
            gradient_norm,
            max_trustregion_radius,
        )

        trustregion_radius = tau * trustregion_radius

    x_candidate = x_candidate - step_size_accepted * gradient_projected
    x_candidate = apply_bounds_to_x_candidate(x_candidate, lower_bounds, upper_bounds)

    return (
        x_candidate,
        f_min,
        trustregion_radius,
        max_trustregion_radius,
    )


def compute_newton_step(model):
    """Compute the newton step."""
    n = model.linear_terms.shape[0]
    diag_min = np.ones(n) * 1e-6
    diag_max = np.ones(n) * 1e6

    diag = np.abs(np.diag(model.square_terms))
    diag_bounded = np.median(np.stack([diag_min, diag, diag_max]), axis=0)

    newton_step = model.linear_terms / diag_bounded

    return newton_step


def update_trustregion_radius_conjugate_gradient(
    f_candidate, predicted_reduction, actual_reduction, x_norm_cg, tr_radius
):
    """Update the trust-region radius based on predicted and actual reduction."""
    eta1 = 1.0e-4
    eta2 = 0.25
    eta3 = 0.50
    eta4 = 0.90

    alpha1 = 0.25
    alpha2 = 0.50
    alpha3 = 1.00
    alpha4 = 2.00
    alpha5 = 4.00

    accept_step = False

    if predicted_reduction < 0 or np.isfinite(predicted_reduction) is False:
        # The predicted reduction has the wrong sign.
        # Reject the just peformed Conjuate Gradient step and start over.
        tr_radius = alpha1 * min(tr_radius, x_norm_cg)
    else:
        if np.isfinite(actual_reduction) is False:
            tr_radius = alpha1 * min(tr_radius, x_norm_cg)
        else:
            if abs(actual_reduction) <= max(1, abs(f_candidate) * EPSILON) and abs(
                predicted_reduction
            ) <= max(1, abs(f_candidate) * EPSILON):
                kappa = 1
            else:
                kappa = actual_reduction / predicted_reduction

            if kappa < eta1:
                tr_radius = alpha1 * min(tr_radius, x_norm_cg)
            else:
                accept_step = True

                # Update the trust-region radius only if the computed step is at the
                # trust radius boundary
                if np.allclose(x_norm_cg, tr_radius):
                    if kappa < eta2:
                        # Marginal bad step
                        tr_radius = alpha2 * tr_radius
                    elif kappa < eta3:
                        # Reasonable step
                        tr_radius = alpha3 * tr_radius
                    elif kappa < eta4:
                        tr_radius = alpha4 * tr_radius
                    else:
                        # Very good step
                        tr_radius = alpha5 * tr_radius

    tr_radius = max(min(tr_radius, 1e10), 1e-10)

    return tr_radius, accept_step


def get_information_on_active_bounds(
    x, lower_bounds, upper_bounds, model, newton_step, step_size_newton
):
    """Return the index set of active bounds."""
    ActiveBounds = namedtuple(
        "ActiveBounds", ["lower", "upper", "fixed", "all", "inactive"]
    )

    x_candidate = np.copy(x)
    x_candidate = x_candidate - step_size_newton * newton_step
    x_candidate_bounded = apply_bounds_to_x_candidate(
        x_candidate, lower_bounds, upper_bounds
    )

    centered_x = x - x_candidate_bounded

    active_lower = np.where((centered_x <= lower_bounds) & (model.linear_terms > 0))[0]
    active_upper = np.where((centered_x >= upper_bounds) & (model.linear_terms < 0))[0]
    active_fixed = np.where(lower_bounds == upper_bounds)[0]
    active_all = reduce(np.union1d, (active_fixed, active_lower, active_upper))
    inactive = np.setdiff1d(np.arange(x.shape[0]), active_all)

    active_bounds_info = ActiveBounds(
        lower=active_lower,
        upper=active_upper,
        fixed=active_fixed,
        all=active_all,
        inactive=inactive,
    )

    return active_bounds_info


def find_hessian_submatrix_where_bounds_inactive(model, active_bounds_info):
    """Find the submatrix of the initial hessian where bounds are inactive."""
    hessian_inactive = model.square_terms[
        active_bounds_info.inactive[:, np.newaxis], active_bounds_info.inactive
    ]

    return hessian_inactive


def check_for_convergence_conjugate_gradient(
    x_candidate,
    f_candidate,
    x_old,
    f_old,
    gradient_candidate,
    model,
    lower_bounds,
    upper_bounds,
    trustregion_radius,
    *,
    ftol_abs,
    ftol_scaled,
    xtol,
    gtol_abs,
    gtol_rel,
    gtol_scaled,
    steptol,
):
    """Check if we have found a solution."""
    direction_fischer_burmeister = _get_fischer_burmeister_direction_vector(
        x_candidate, gradient_candidate, lower_bounds, upper_bounds
    )
    gradient_norm = np.linalg.norm(direction_fischer_burmeister)

    if gradient_norm < gtol_abs:
        converged = True
    elif f_candidate != 0 and abs(gradient_norm / f_candidate) < gtol_rel:
        converged = True
    elif gradient_norm / np.linalg.norm(model.linear_terms) < gtol_scaled:
        converged = True
    elif abs(f_old - f_candidate) < ftol_abs:
        converged = True
    elif (f_old - f_candidate) / max(abs(f_old), abs(f_candidate), 1) < ftol_scaled:
        converged = True
    elif np.max(np.abs(x_old - x_candidate)) < xtol:
        converged = True
    elif trustregion_radius < steptol:
        converged = True
    else:
        converged = False

    return converged


def check_for_convergence_steepest_descent(
    x_candidate,
    f_candidate,
    gradient_candidate,
    model,
    lower_bounds,
    upper_bounds,
    gtol_abs,
    gtol_rel,
    gtol_scaled,
):
    """Check if we have found a solution."""
    direction_fischer_burmeister = _get_fischer_burmeister_direction_vector(
        x_candidate, gradient_candidate, lower_bounds, upper_bounds
    )
    gradient_norm = np.linalg.norm(direction_fischer_burmeister)

    if gradient_norm < gtol_abs:
        converged = True
    elif f_candidate != 0 and abs(gradient_norm / f_candidate) < gtol_rel:
        converged = True
    elif gradient_norm / np.linalg.norm(model.linear_terms) < gtol_scaled:
        converged = True
    else:
        converged = False

    return converged


def apply_bounds_to_x_candidate(x, lower_bounds, upper_bounds, bound_tol=0):
    """Apply upper and lower bounds to the candidate vector."""
    x = np.where(x <= lower_bounds + bound_tol, lower_bounds, x)
    x = np.where(x >= upper_bounds - bound_tol, upper_bounds, x)

    return x


def evaluate_model_criterion(
    x,
    gradient,
    hessian,
):
    """Evaluate the criterion function value of the main model.

    Args:
        x (np.ndarray): Parameter vector of shape (n,).
        gradient (np.ndarray): Gradient of shape (n,) for which the main model
            shall be evaluated.
        hessian (np.ndarray): Hessian of shape (n, n) for which the main model
            shall be evaulated.

    Returns:
        (float): Criterion value of the main model.
    """
    return np.dot(gradient, x) + 0.5 * np.dot(np.dot(x, hessian), x)


def evaluate_model_gradient(x, model):
    """Evaluate the derivative of the main model.

    Args:
       main_model (namedtuple): Named tuple containing the parameters of the
            main model, i.e.:
            - "linear_terms", a np.ndarray of shape (n,) and
            - "square_terms", a np.ndarray of shape (n,n).

    Returns:
        (np.ndarray): Derivative of the main model of shape (n,).
    """
    return model.linear_terms + np.dot(model.square_terms, x)


def _apply_bounds_to_conjugate_gradient_step(
    x,
    lower_bounds,
    upper_bounds,
    active_bounds_info,
    cg_step_direction=None,
):
    """Apply lower and upper bounds to the Conjugate Gradient step direction."""
    n = x.shape[0]

    if cg_step_direction is None:
        cg_step_direction = np.zeros(n)

    if active_bounds_info.lower.size > 0:
        x_active_lower = x[active_bounds_info.lower]
        lower_bound_active = lower_bounds[active_bounds_info.lower]

        cg_step_direction[active_bounds_info.lower] = (
            lower_bound_active - x_active_lower
        )

    if active_bounds_info.upper.size > 0:
        x_active_upper = x[active_bounds_info.upper]
        upper_bound_active = upper_bounds[active_bounds_info.upper]

        cg_step_direction[active_bounds_info.upper] = (
            upper_bound_active - x_active_upper
        )

    if active_bounds_info.fixed.size > 0:
        cg_step_direction[active_bounds_info.fixed] = 0

    return cg_step_direction


def _update_tau_and_max_trustregion_radius_steepest_descent(
    trustregion_radius, prered, actred, gradient_norm, max_radius
):
    """Update tau and the new maximum for the trust-region radius."""
    mu1 = 0.35
    mu2 = 0.50
    gamma1 = 0.0625
    gamma2 = 0.5
    gamma3 = 2.0
    gamma4 = 5.0
    theta = 0
    epsilon = np.finfo(float).eps ** (2 / 3)

    if abs(actred) <= epsilon and abs(prered) <= epsilon:
        kappa = 1
    else:
        kappa = actred / prered

    tau_1 = (
        theta
        * gradient_norm
        * trustregion_radius
        / (theta * gradient_norm * trustregion_radius + (1 - theta) * prered - actred)
    )
    tau_2 = (
        theta
        * gradient_norm
        * trustregion_radius
        / (theta * gradient_norm * trustregion_radius - (1 + theta) * prered + actred)
    )

    tau_min = min(tau_1, tau_2)
    tau_max = max(tau_1, tau_2)

    if abs(kappa - 1) <= mu1:
        # Great agreement
        max_radius = max(max_radius, trustregion_radius)

        if tau_max < 1:
            tau = gamma3
        elif tau_max > gamma4:
            tau = gamma4
        else:
            tau = tau_max

    elif abs(kappa - 1) <= mu2:
        # Good agreement
        max_radius = max(max_radius, trustregion_radius)

        if tau_max < gamma2:
            tau = gamma2
        elif tau_max > gamma3:
            tau = gamma3
        else:
            tau = tau_max

    else:
        # Not good agreement
        if tau_min > 1:
            tau = gamma2
        elif tau_max < gamma1:
            tau = gamma1
        elif (tau_min < gamma1) and (tau_max >= 1):
            tau = gamma1
        elif (
            (tau_1 >= gamma1) and (tau_1 < 1.0) and ((tau_2 < gamma1) or (tau_2 >= 1.0))
        ):
            tau = tau_1
        elif (
            (tau_2 >= gamma1) and (tau_2 < 1.0) and ((tau_1 < gamma1) or (tau_2 >= 1.0))
        ):
            tau = tau_2
        else:
            tau = tau_max

    return tau, max_radius


def _get_fischer_burmeister_direction_vector(x, gradient, lower_bounds, upper_bounds):
    """Compute the constrained direction vector via the Fischer-Burmeister function."""
    fischer_vec = np.vectorize(_get_fischer_burmeister_scalar)

    fischer_burmeister = reduce(
        fischer_vec, (upper_bounds - x, -gradient, x - lower_bounds)
    )
    direction = np.where(
        lower_bounds == upper_bounds, lower_bounds - x, fischer_burmeister
    )

    return direction


def _get_fischer_burmeister_scalar(a, b):
    """Get the value of the Fischer-Burmeister function for two scalar inputs.

    This method was suggested by Bob Vanderbei. Since the Fischer-Burmeister
    is symmetric, the order of the scalar inputs does not matter.

    Args:
        a (float): First input.
        b (float): Second input.

    Returns:
        float: Value of the Fischer-Burmeister function for inputs a and b.
    """
    if a + b <= 0:
        fischer_burmeister = np.sqrt(a**2 + b**2) - (a + b)
    else:
        fischer_burmeister = -2 * a * b / (np.sqrt(a**2 + b**2) + (a + b))

    return fischer_burmeister



"""Collection of solvers for a quadratic trust-region subproblem."""
from collections import namedtuple
from copy import copy

import numpy as np
from estimagic.optimization._trustregion_bounded_newton_quadratic import (
    apply_bounds_to_x_candidate,
)
from estimagic.optimization._trustregion_bounded_newton_quadratic import (
    check_for_convergence_conjugate_gradient,
)
from estimagic.optimization._trustregion_bounded_newton_quadratic import (
    compute_conjugate_gradient_step,
)
from estimagic.optimization._trustregion_bounded_newton_quadratic import (
    compute_predicted_reduction_from_conjugate_gradient_step,
)
from estimagic.optimization._trustregion_bounded_newton_quadratic import (
    take_preliminary_steepest_descent_step_and_check_for_solution,
)
from estimagic.optimization._trustregion_bounded_newton_quadratic import (
    update_trustregion_radius_conjugate_gradient,
)
from estimagic.optimization._trustregion_gqtpar_quadratic import (
    add_lambda_and_factorize_hessian,
)
from estimagic.optimization._trustregion_gqtpar_quadratic import (
    check_for_interior_convergence_and_update,
)
from estimagic.optimization._trustregion_gqtpar_quadratic import (
    find_new_candidate_and_update_parameters,
)
from estimagic.optimization._trustregion_gqtpar_quadratic import (
    get_initial_guess_for_lambdas,
)
from estimagic.optimization._trustregion_gqtpar_quadratic import (
    update_lambdas_when_factorization_unsuccessful,
)


def minimize_bntr_quadratic(
    model,
    lower_bounds,
    upper_bounds,
    *,
    maxiter,
    maxiter_steepest_descent,
    step_size_newton,
    ftol_abs,
    ftol_scaled,
    xtol,
    gtol_abs,
    gtol_rel,
    gtol_scaled,
    steptol
):
    """Minimize a bounded trust-region subproblem via Newton Conjugate Gradient method.

    The BNTR (Bounded Newton Trust Rregion) algorithm uses an active-set approach
    to solve the symmetric system of equations:

        Hessian @ x = - gradient

    only for the inactive parameters of x that lie within the bounds. The active-set
    estimation employed here is based on Bertsekas (:cite:`Bertsekas1982`).

    In the main loop, BNTR globalizes the Newton step using a trust-region method
    based on the predicted versus actual reduction in the criterion function.
    The trust-region radius is increased only if the accepted step is at the
    trust-region boundary.


    Args:
        model (namedtuple): Named tuple containing the parameters of the
            main model, i.e.:
            - "linear_terms", a np.ndarray of shape (n,) and
            - "square_terms", a np.ndarray of shape (n,n).
        lower_bounds (np.ndarray): Lower bound on parameter vector x.
            Must have same length as the initial guess of the
            parameter vector. Equal to -1 if not provided by the user.
        upper_bounds (np.ndarray): Upper bounds on parameter vector x.
        maxiter (int): Maximum number of iterations. If reached, terminate.
        maxiter_steepest_descent (int): Maximum number of steepest descent iterations
            to perform when the trust-region subsolver BNTR is used.
        step_size_newton (float): Parameter to scale the size of the newton step
            before the preliminary steepest descent step is taken.
        ftol_abs (float): Convergence tolerance for the absolute difference
            between f(k+1) - f(k).
        ftol_scaled (float): Convergence tolerance for the scaled difference
            between f(k+1) - f(k).
        xtol_sub (float): Convergence tolerance for the absolute difference
            between max(x(k+1) - x(k)).
        gtol_abs_sub (float): Convergence tolerance for the absolute gradient norm.
        gtol_rel_sub (float): Convergence tolerance for the relative gradient norm.
        gtol_scaled_sub (float): Convergence tolerance for the scaled gradient norm.

    Returns:
        (dict): Result dictionary containing the following keys:
            - "x" (np.ndarray): Solution vector of the subproblem of shape (n,)
            - "criterion" (float): Minimum function value associated with the
                solution.
            - n_iterations (int): Number of iterations the algorithm ran before
                termination.
            - "success" (bool): Boolean indicating whether a solution has been found
                before reaching maxiter.
    """
    x_candidate = np.zeros(model.linear_terms.shape[0])

    (
        x_candidate,
        f_candidate,
        gradient_candidate,
        hessian_inactive,
        trustregion_radius,
        active_bounds_info,
        converged,
    ) = take_preliminary_steepest_descent_step_and_check_for_solution(
        x_candidate,
        model,
        lower_bounds,
        upper_bounds,
        maxiter_steepest_descent,
        step_size_newton,
        gtol_abs,
        gtol_rel,
        gtol_scaled,
    )

    for _niter in range(maxiter):
        x_old = np.copy(x_candidate)
        f_old = copy(f_candidate)

        accept_step = False

        while accept_step is False and converged is False:
            (
                x_inactive_conjugate_gradient,
                gradient_inactive_conjugate_gradient,
                x_norm_conjugate_gradient,
                conjugate_gradient_step,
            ) = compute_conjugate_gradient_step(
                x_candidate,
                gradient_candidate,
                hessian_inactive,
                lower_bounds,
                upper_bounds,
                active_bounds_info,
                trustregion_radius,
            )

            # Temporarily accept the step and project it into the bounds
            x_candidate = x_candidate - conjugate_gradient_step
            x_candidate = apply_bounds_to_x_candidate(
                x_candidate, lower_bounds, upper_bounds
            )
            f_candidate = evaluate_model_criterion(
                x_candidate, model.linear_terms, model.square_terms
            )

            predicted_reduction = (
                compute_predicted_reduction_from_conjugate_gradient_step(
                    x_inactive_conjugate_gradient,
                    gradient_inactive_conjugate_gradient,
                    conjugate_gradient_step,
                    gradient_candidate,
                    hessian_inactive,
                    active_bounds_info,
                )
            )

            actual_reduction = f_old - f_candidate

            trustregion_radius_old = copy(trustregion_radius)

            (
                trustregion_radius,
                accept_step,
            ) = update_trustregion_radius_conjugate_gradient(
                f_candidate,
                predicted_reduction,
                actual_reduction,
                x_norm_conjugate_gradient,
                trustregion_radius,
            )

            if accept_step:
                gradient_candidate = evaluate_model_gradient(x_candidate, model)
            else:
                x_candidate = np.copy(x_old)
                f_candidate = np.copy(f_old)

                if trustregion_radius_old == trustregion_radius:
                    break

            converged = check_for_convergence_conjugate_gradient(
                x_candidate,
                f_candidate,
                x_old,
                f_old,
                gradient_candidate,
                model,
                lower_bounds,
                upper_bounds,
                trustregion_radius,
                ftol_abs=ftol_abs,
                ftol_scaled=ftol_scaled,
                xtol=xtol,
                gtol_abs=gtol_abs,
                gtol_rel=gtol_rel,
                gtol_scaled=gtol_scaled,
                steptol=steptol,
            )

        if converged is True:
            break

    result = {
        "x": x_candidate,
        "criterion": f_candidate,
        "n_iterations": _niter,
        "success": converged,
    }

    return result


def minimize_gqtpar_quadratic(model, *, k_easy=0.1, k_hard=0.2, maxiter=200):
    """Solve the quadratic trust-region subproblem via nearly exact iterative method.

    This subproblem solver is mainly based on Conn et al. (2000) "Trust region methods"
    (:cite:`Conn2000`), pp. 169-200.

    But ideas from Nocedal and Wright (2006) "Numerical optimization"
    (:cite:`Nocedal2006`), pp. 83-91, who implement a similar algorithm,
    were also used.

    The original algorithm was developed by More and Sorensen (1983) (:cite:`More1983`)
    and is known as "GQTPAR".

    The direction vector ``p*`` is a global solution to the quadratic subproblem:

        min f + g p + 0.5 * p H p,

        s.t. norm(p) <= trustregion_radius

        if and only if norm(``p*``) <= trustregion radius and there is a scalar
        lambda >= 0, such that:

    1) (H + lambda * I(n)) p* = -g
    2) lambda (trustregion_radius - norm(p*)) = 0
    3) H + lambda * I is positive definite

    where g denotes the gradient vector and H the hessian matrix of the main model,
    respectively.

    k_easy and k_hard are stopping criteria for the iterative subproblem solver.
    See pp. 194-197 in :cite:`Conn2000` for a more detailed description.

    Args:
        main_model (namedtuple): Named tuple containing the parameters of the
            main model, i.e.:
            - "linear_terms", a np.ndarray of shape (n,) and
            - "square_terms", a np.ndarray of shape (n,n).
        trustregion_radius (float): Trustregion radius, often referred to as delta.
        k_easy (float): topping criterion for the "easy" case.
        k_hard (float): Stopping criterion for the "hard" case.
        maxiter (int): Maximum number of iterations to perform. If reached,
            terminate.

    Returns:
        (dict): Result dictionary containing the following keys:
            - "x" (np.ndarray): Solution vector of the subproblem of shape (n,)
            - "criterion" (float): Minimum function value associated with the
                solution.
    """
    # Small floating point number signaling that for vectors smaller
    # than that backward substituition is not reliable.
    # See Golub, G. H., Van Loan, C. F. (2013), "Matrix computations", p.165.
    zero_threshold = (
        model.square_terms.shape[0]
        * np.finfo(float).eps
        * np.linalg.norm(model.square_terms, np.Inf)
    )
    stopping_criteria = {
        "k_easy": k_easy,
        "k_hard": k_hard,
    }

    HessianInfo = namedtuple(
        "HessianInfo", ["hessian_plus_lambda", "upper_triangular", "already_factorized"]
    )
    hessian_info = HessianInfo(
        hessian_plus_lambda=None,
        upper_triangular=None,
        already_factorized=False,
    )

    gradient_norm = np.linalg.norm(model.linear_terms)
    lambdas = get_initial_guess_for_lambdas(model)

    converged = False

    for _niter in range(maxiter):

        if hessian_info.already_factorized is True:
            hessian_info = hessian_info._replace(already_factorized=False)
        else:
            hessian_info, factorization_info = add_lambda_and_factorize_hessian(
                model, hessian_info, lambdas
            )

        if factorization_info == 0 and gradient_norm > zero_threshold:
            (
                x_candidate,
                hessian_info,
                lambdas,
                converged,
            ) = find_new_candidate_and_update_parameters(
                model,
                hessian_info,
                lambdas,
                stopping_criteria,
                converged,
            )

        elif factorization_info == 0 and gradient_norm <= zero_threshold:
            x_candidate, lambdas, converged = check_for_interior_convergence_and_update(
                x_candidate,
                hessian_info,
                lambdas,
                stopping_criteria,
                converged,
            )

        else:
            lambdas = update_lambdas_when_factorization_unsuccessful(
                hessian_info,
                lambdas,
                factorization_info,
            )

        if converged is True:
            break

    f_min = evaluate_model_criterion(
        x_candidate, model.linear_terms, model.square_terms
    )

    result = {
        "x": x_candidate,
        "criterion": f_min,
        "n_iterations": _niter,
        "success": converged,
    }

    return result


def evaluate_model_criterion(
    x,
    gradient,
    hessian,
):
    """Evaluate the criterion function value of the main model.

    Args:
        x (np.ndarray): Parameter vector of shape (n,).
        gradient (np.ndarray): Gradient of shape (n,) for which the main model
            shall be evaluated.
        hessian (np.ndarray): Hessian of shape (n, n) for which the main model
            shall be evaulated.

    Returns:
        (float): Criterion value of the main model.
    """
    return np.dot(gradient, x) + 0.5 * np.dot(np.dot(x, hessian), x)


def evaluate_model_gradient(x, model):
    """Evaluate the derivative of the main model.

    Args:
       main_model (namedtuple): Named tuple containing the parameters of the
            main model, i.e.:
            - "linear_terms", a np.ndarray of shape (n,) and
            - "square_terms", a np.ndarray of shape (n,n).

    Returns:
        (np.ndarray): Derivative of the main model of shape (n,).
    """
    return model.linear_terms + np.dot(model.square_terms, x)



try:
    from better_exceptions import format_exception
except ImportError:
    from traceback import format_exception
import sys


class TableExistsError(Exception):
    pass


class StopOptimizationError(Exception):
    def __init__(self, message, current_status):
        super().__init__(message)
        self.message = message
        self.current_status = current_status

    def __reduce__(self):
        """Taken from here: https://tinyurl.com/y6eeys2f"""
        return (StopOptimizationError, (self.message, self.current_status))


def get_traceback():
    tb = format_exception(*sys.exc_info())
    if isinstance(tb, list):
        tb = "".join(tb)
    return tb


INVALID_INFERENCE_MSG = (
    "Taking the inverse of the information matrix failed. Only ever use this "
    "covariance matrix or standard errors based on it for diagnostic purposes, not for "
    "drawing conclusions."
)


INVALID_SENSITIVITY_MSG = (
    "Taking inverse failed during the calculation of sensitvity measures. Interpret "
    "them with caution."
)
